from dungeonsheets import spells, weapons
from dungeonsheets.features.features import Feature, FeatureSelector


# Features
class EldritchInvocation(Feature):
    """In your study of occult lore, you have unearthed eldritch invocations,
    fragments of forbidden knowledge that imbue you with an abiding magical
    ability.

    At 2nd level, you gain two eldritch invocations of your choice. Your
    invocation options are detailed at the end of the class description. When
    you gain certain warlock levels, you gain additional invocations of your
    choice, as shown in the Invocations Known column of the Warlock table.

    Additionally, when you gain a level in this class, you can choose one of
    the invocations you know and replace it with another invocation that you
    could learn at that level.

    """

    name = "Eldritch Invocations"
    source = "Warlock"


class PactOfTheChain(Feature):
    """You learn the find familiar spell and can cast it as a ritual. The spell
    doesn't count against your number of spells known.

    When you cast the spell, you can choose one of the normal forms for your
    familiar or one of the following special forms: imp, pseudodragon, quasit,
    or sprite. Additionally, when you take the Attack action, you can forgo one
    of your own attacks to allow your familiar to make one attack of its own.

    """

    name = "Pact of the Chain"
    source = "Warlock"
    spells_known = spells_prepared = (spells.FindFamiliar,)


class PactOfTheBlade(Feature):
    """You can use your action to create a pact weapon in your empty hand. You can
    choose the form that this melee weapon takes each time you create it (see
    chapter 5 for weapon options). You are proficient with it while you wield
    it. This weapon counts as magical for the purpose o f overcoming resistance
    and immunity to nonmagical attacks and damage.

    Your pact weapon disappears if it is more than 5 feet away from you for 1
    minute or more. It also disappears if you use this feature again, if you
    dismiss the weapon (no action required), or if you die.

    You can transform one magic weapon into your pact weapon by performing a
    special ritual while you hold the weapon. You perform the ritual over the
    course of 1 hour, which can be done during a short rest. You can then
    dismiss the weapon, shunting it into an extradimensional space, and it
    appears whenever you create your pact weapon thereafter. You can't affect
    an artifact or a sentient weapon in this way. The weapon ceases being your
    pact weapon if you die, if you perform the 1-hour ritual on a different
    weapon, or if you use a 1-hour ritual to break your bond to it. The weapon
    appears at your feet if it is in the extradimensional space when the bond
    breaks.

    """

    name = "Pact of the Blade"
    source = "Warlock"


class PactOfTheTome(Feature):
    """Your patron gives you a grimoire called a Book of Shadows. When you gain
    this feature, choose three cantrips from any class's spell list. While the
    book is on your person, you can cast those cantrips at will. They don't
    count against your number of cantrips known.

    If you lose your Book of Shadows, you can perform a 1-hour ceremony to
    receive a replacement from your patron. This ceremony can be performed
    during a short or long rest, and it destroys the previous book. The book
    turns to ash when you die.

    """

    name = "Pact of the Tome"
    source = "Warlock"


class PactBoon(FeatureSelector):
    """Select a Pact Boon by choosing in feature_choices:

    pact of the chain

    pact of the blade

    pact of the tome

    """

    options = {
        "chain": PactOfTheChain,
        "pact of the chain": PactOfTheChain,
        "blade": PactOfTheBlade,
        "pact of the blade": PactOfTheBlade,
        "tome": PactOfTheTome,
        "pact of the tome": PactOfTheTome,
    }
    name = "Pact Boon (Select One)"
    source = "Warlock"


class MysticArcanum(Feature):
    """At 11th level, your patron bestows upon you a magical secret called an
    arcanum. Choose one 6th-level spell from the warlock spell list as this
    arcanum.

    You can cast your arcanum spell once without expending a spell slot. You
    must finish a long rest before you can do so again.

    At higher levels, you gain more warlock spells of your choice that can be
    cast in this way: one 7th-level spell at 13th level, one 8th-level spell at
    15th level, and one 9th-level spell at 17th level. You regain all uses of
    your Mystic Arcanum when you finish a long rest.

    """

    name = "Mystic Arcanum"
    source = "Warlock"


class EldritchMaster(Feature):
    """At 20th level, you can draw on your inner reserve of mystical power while
    entreating your patron to regain expended spell slots. You can spend 1
    minute entreating your patron for aid to regain all your expended spell
    slots from your Pact Magic feature. Once you regain spell slots with this
    feature, you must finish a long rest before you can do so again.

    """

    name = "Eldritch Master"
    source = "Warlock"


# The Archfey
class FeyPresence(Feature):
    """Starting at 1st level, your patron bestows upon you the ability to project
    the beguiling and fearsome presence of the fey. As an action, you can cause
    each creature in a 10-foot cube originating from you to make a Wisdom
    saving throw against your warlock spell save DC.

    The creatures that fail their saving throws are all charmed or frightened
    by you (your choice) until the end of your next turn. Once you use this
    feature, you can't use it again until you finish a short or long rest.

    """

    name = "Fey Presence"
    source = "Warlock (Archfey Patron)"


class MistyEscape(Feature):
    """Starting at 6th level, you can vanish in a puff of mist in response to
    harm. When you take damage, you can use your reaction to turn invisible and
    teleport up to 60 feet to an unoccupied space you can see. You remain
    invisible until the start of your next turn or until you attack or cast a
    spell. Once you use this feature, you can't use it again until you finish a
    short or long rest

    """

    name = "Misty Escape"
    source = "Warlock (Archfey Patron)"


class BeguilingDefenses(Feature):
    """Beginning at 10th level, your patron teaches you how to turn the
    mind-affecting magic of your enemies against them. You are immune to being
    charmed, and when another creature attempts to charm you, you can use your
    reaction to attempt to turn the charm back on that creature. The creature
    must succeed on a Wisdom saving throw against your warlock spell save DC or
    be charmed by you for 1 minute or until the creature takes any damage.

    """

    name = "Beguiling Defenses"
    source = "Warlock (Archfey Patron)"


class DarkDelirium(Feature):
    """Starting at 14th level, you can plunge a creature into an illusory
    realm. As an action, choose a creature that you can see within 60 feet of
    you. It must make a Wisdom saving throw against your warlock spell save
    DC. On a failed save, it is charmed or frightened by you (your choice) for
    1 minute or until your concentration is broken (as if you are concentrating
    on a spell).

    This effect ends early if the creature takes any damage. Until this
    illusion ends, the creature thinks it is lost in a misty realm, the
    appearance of which you choose. The creature can see and hear only itself,
    you, and the illusion. You must finish a short or long rest before you can
    use this feature again.

    """

    name = "Dark Delirium"
    source = "Warlock (Archfey Patron)"


# The Fiend Patron
class DarkOnesBlessing(Feature):
    """Starting at 1st level, when you reduce a hostile creature to 0 hit points,
    you gain temporary hit points equal to your Charisma modifier + your
    warlock level (minimum of 1)

    """

    _name = "Dark One's Blessing"
    source = "Warlock (The Fiend Patron)"

    @property
    def name(self):
        level = self.owner.Warlock.level
        mod = self.owner.charisma.modifier
        return self._name + " ({:d} HP)".format(level + mod)


class DarkOnesOwnLuck(Feature):
    """Starting at 6th level, you can call on your patron to alter fate in your
    favor. When you make an ability check or a saving throw, you can use this
    feature to add a d10 to your roll. You can do so after seeing the initial
    roll but before any o f the roll's effects occur.

    Once you use this feature, you can't use it again until you finish a short
    or long rest.

    """

    name = "Dark One's Own Luck"
    source = "Warlock (The Fiend Patron)"


class FiendishResilience(Feature):
    """Starting at 10th level, you can choose one damage type when you finish a
    short or long rest. You gain resistance to that damage type until you
    choose a different one with this feature. Damage from magical weapons or
    silver weapons ignores this resistance.

    """

    name = "Fiendish Resilience"
    source = "Warlock (The Fiend Patron)"


class HurlThroughHell(Feature):
    """Starting at 14th level, when you hit a creature with an attack, you can use
    this feature to instantly transport the target through the lower
    planes. The creature disappears and hurtles through a nightmare landscape.

    At the end of your next turn, the target returns to the space it previously
    occupied, or the nearest unoccupied space. If the target is not a fiend, it
    takes 10d10 psychic damage as it reels from its horrific experience.

    Once you use this feature, you can't use it again until you finish a long
    rest.

    """

    name = "Hurl Through Hell"
    source = "Warlock (The Fiend Patron)"


# Great Old One
class AwakenedMind(Feature):
    """Starting at 1st level, your alien knowledge gives you the ability to touch
    the minds of other creatures. You can communicate telepathically with any
    creature you can see within 30 feet of you. You don't need to share a
    language with the creature for it to understand your telepathic utterances,
    but the creature must be able to understand at least one language

    """

    name = "Awakened Mind"
    source = "Warlock (Great Old One Patron)"


class EntropicWard(Feature):
    """At 6th level, you learn to magically ward yourself against attack and to
    turn an enemy's failed strike into good luck for yourself. When a creature
    makes an attack roll against you, you can use your reaction to impose
    disadvantage on that roll. If the attack misses you, your next attack roll
    against the creature has advantage if you make it before the end of your
    next turn. Once you use this feature, you can't use it again until you
    finish a short or long rest.

    """

    name = "Entropic Ward"
    source = "Warlock (Great Old One Patron)"


class ThoughtShield(Feature):
    """Starting at 10th level, your thoughts can't be read by telepathy or other
    means unless you allow it. You also have resistance to psychic damage, and
    whenever a creature deals psychic damage to you, that creature takes the
    same amount of damage that you do

    """

    name = "Thought Shield"
    source = "Warlock (Great Old One Patron)"


class CreateThrall(Feature):
    """At 14th level, you gain the ability to infect a humanoid's mind with the
    alien magic of your patron. You can use your action to touch an
    incapacitated humanoid. That creature is then charmed by you until a remove
    curse spell is cast on it, the charmed condition is removed from it, or you
    use this feature again. You can communicate telepathically with the charmed
    creature as long as the two of you are on the same plane of existence

    """

    name = "Create Thrall"
    source = "Warlock (Great Old One Patron)"


# Undying Patron
class AmongTheDead(Feature):
    """Starting at 1st level, you learn the spare the dying cantrip, which counts
    as a warlock cantrip for you. You also have advantage on saving throws
    against any disease. Additionally, undead have difficulty harming you. If
    an undead targets you directly with an attack or a harmful spell, that
    creature must make a Wisdom saving throw against your spell save DC (an
    undead needn't make the save when it includes you in an area effect, such
    as the explosion of fireball). On a failed save, the creature must choose a
    new target or forfeit targeting someone instead of you, potentially wasting
    the attack or spell. On a successful save, the creature is immune to this
    effect for 24 hours. An undead is also immune to this effect for 24 hours
    if you target it with an attack or a harmful spell.

    """

    name = "Among the Dead"
    source = "Warlock (The Undying Patron)"
    spells_known = spells_prepared = (spells.SpareTheDying,)


class DefyDeath(Feature):
    """Starting at 6th level, you can give yourself vitality when you cheat death
    or when you help someone else cheat it. You can regain hit points equal to
    ld8 +your Constitution modifier (minimum of 1 hit point) when you succeed
    on a death saving throw or when you stabilize a creature with spare the
    dying. Once you use this feature, you can't use it again until you finish a
    long rest

    """

    name = "Defy Death"
    source = "Warlock (The Undying Patron)"


class UndyingNature(Feature):
    """Beginning at 10th level , you can hold your breath indefinitely, and you
    don't require food, water, or sleep, although you still require rest to
    reduce exhaustion and still benefit from finishing short and long rests. In
    addition, you age at a slower rate. For every 10 years that pass, your body
    ages only 1 year, and you are immune to being magically aged

    """

    name = "Undying Nature"
    source = "Warlock (The Undying Patron)"


class IndestructibleLife(Feature):
    """When you reach 14th level, you partake of some of the true secrets
    of the Undying. On your turn, you can use a bonus action to regain
    hit points equal to 1d8 + your warlock level. Additionally, if you
    put a severed body part of yours back in place when you use this
    feature, the part reattaches. Once you use this feature, you can't
    use it again until you finish a short or long rest.

    """

    name = "Indestructible Life"
    source = "Warlock (The Undying Patron)"


# The Celestial
class HealingLight(Feature):
    """At lst level, you gain the ability to channel celestial energy to heal
    wounds. You have a pool of d6s that you spend to fuel this healing. The
    number of dice in the pool equals 1 + your warlock level. As a bonus
    action, you can heal one creature you can see within 60 feet of you,
    spending dice from the pool. The maximum number of dice you can spend at
    once equals your Charisma modifier (minimum of one die). Roll the
    dice you spend, add them together, and restore a number of hit points equal
    to the total. Your pool regains all expended dice when you finish a long
    rest

    """

    _name = "Healing Light"
    source = "Warlock (The Celestial Patron)"

    @property
    def name(self):
        num = 1 + self.owner.Warlock.level
        return self._name + " ({:d}d6/LR)".format(num)


class RadiantSoul(Feature):
    """Starting at 6th level, your link to the Celestial allows you to serve as a
    conduit for radiant energy. You have resistance to radiant damage, and when
    you cast a spell that deals radiant or fire damage, you can add your
    Charisma modifier to one radiant or fire damage roll of that spell against
    one of its targets.

    """

    name = "Radiant Soul"
    source = "Warlock (The Celestial Patron)"


class CelestialResilience(Feature):
    """Starting at 10th level, you gain temporary hit points whenever you finish a
    short or long rest. These tempo- rary hit points equal your warlock level +
    your Charisma modifier. Additionally, choose up to five creatures you can
    see at the end of the rest. Those creatures each gain temporary hit points
    equal to half your warlock level + your Charisma modifier

    """

    _name = "Celestial Resilience"
    source = "Warlock (The Celestial Patron)"


class SearingVengeance(Feature):
    """Starting at 14th level, the radiant energy you channel allows you to resist
    death. When you have to make a death saving throw at the start of your
    turn, you can instead spring back to your feet with a burst of radiant
    energy. You regain hit points equal to half your hit point maximum, and
    then you stand up if you so choose. Each creature of your choice that is
    within 30 feet of you takes radiant damage equal to 2d8 + your Charisma
    modifier, and it is blinded until the end of the current turn. Once you use
    this feature, you can't use it again until you finish a long rest.

    """

    name = "Searing Vengeance"
    source = "Warlock (The Celestial Patron)"


# Hexblade
class HexbladesCurse(Feature):
    """Starting at lst level, you gain the ability to place a bale- ful curse on
    someone. As a bonus action, choose one creature you can see within 30 feet
    of you. The target is cursed for 1 minute. The curse ends early if the
    target dies, you die, or you are incapacitated. Until the curse ends, you
    gain the following benefits:

    - You gain a bonus to damage rolls against the cursed target. The
      bonus equals your proficiency bonus.
    - Any attack roll you make against the cursed target is a critical
      hit on a roll of 19 or 20 on the d20.
    - If the cursed target dies, you regain hit points equal to your
      warlock level + your Charisma modifier (minimum of 1 hit point).

    You can't use this feature again until you finish a short or long rest.

    """

    name = "Hexblades Curse"
    source = "Warlock (Hexblade)"


class HexWarrior(Feature):
    """At lst level, you acquire the training necessary to effectively arm
    yourself for battle. You gain proficiency with medium armor, shields, and
    martial weapons.

    The influence of your patron also allows you to mystically channel your
    will through a particular weapon. Whenever you finish a long rest, you can
    touch one weapon that you are proficient with and that lacks the two-handed
    property. When you attack with that weapon, you can use your Charisma
    modifier, instead of Strength or Dexterity, for the attack and damage
    rolls. This benefit lasts until you finish a long rest. If you later gain
    the Pact of the Blade feature, this benefit extends to every pact weapon
    you conjure with that feature, no matter the weapon's type

    """

    name = "Hex Warrior"
    source = "Warlock (Hexblade)"

    def weapon_func(self, weapon: weapons.Weapon, **kwargs):
        """
        Swap the weapon's attack bonus modifier for Charisma if
        it is higher than STR/DEX bonus
        """
        if weapon.is_finesse:
            abils = {
                "strength": self.owner.strength.modifier,
                "dexterity": self.owner.dexterity.modifier,
                "charisma": self.owner.charisma.modifier,
            }
        else:
            abils = {
                weapon.ability: getattr(self.owner, weapon.ability).modifier,
                "charisma": self.owner.charisma.modifier,
            }
        weapon.ability = max(abils, key=abils.get)


class AccursedSpecter(Feature):
    """Starting at 6th level, you can curse the soul of a person you slay,
    temporarily binding it to your service. When you slay a humanoid, you can
    cause its Spirit to rise from its corpse as a specter, the statistics for
    which are in the Monster Manual. When the specter appears, it gains
    temporary hit points equal to halfyour warlock level. Roll initiative for
    the specter, which has its own turns. It obeys your verbal commands, and it
    gains a special bonus to its attack rolls equal to your Charisma modifier
    (minimum of +0).

    The specter remains in your service until the end of your next long rest,
    at which point it vanishes to the afterlife.

    Once you bind a specter with this feature, you can't use the feature again
    until you finish a long rest.

    """

    name = "Accursed Specter"
    source = "Warlock (Hexblade)"


class ArmorOfHexes(Feature):
    """At 10th level, your hex grows more powerful. If the target cursed by your
    Hexblade's Curse hits you with an attack roll, you can use your reaction to
    roll a d6. On a 4 or higher, the attack instead misses you, regardless of
    its roll.

    """

    name = "Armor of Hexes"
    source = "Warlock (Hexblade)"


class MasterOfHexes(Feature):
    """Starting at 14th level, you can spread your Hexblade's Curse from a slain
    creature to another creature. When the creature cursed by your Hexblade's
    Curse dies, you can apply the curse to a different creature you can see
    within 30 feet of you, provided you aren't incapacitated. When you apply
    the curse in this way, you don't regain hit points from the death of the
    previously cursed creature.

    """

    name = "Master of Hexes"
    source = "Warlock (Hexblade)"


# All Invocations
class Invocation(Feature):
    """
    A generic Eldritch Invocation. Add details in features/warlock.py
    """

    name = "Unnamed Invocation"
    source = "Warlock (Eldritch Invocations)"
    at_will_spells = ()

    def cast_spell_at_will(self, spell):
        s = spell()
        s.level = 0
        if "M" in s.components:
            c = list(s.components)
            c.remove("M")
            s.components = tuple(c)
        self.spells_known += (s,)
        self.spells_prepared += (s,)

    def __init__(self, owner):
        super().__init__(owner)
        for s in self.at_will_spells:
            self.cast_spell_at_will(s)


# PHB
class AgonizingBlast(Invocation):
    """When you cast eldritch blast, add your Charisma modifier to the damage it
    deals on a hit.

    """

    name = "Agonizing Blast"


class ArmorOfShadows(Invocation):
    """You can cast mage armor on yourself at will, without expending a spell slot
    or material components

    """

    name = "Armor of Shadows"
    at_will_spells = (spells.MageArmor,)


class AscendantStep(Invocation):
    """You can cast levitate on yourself at will, without expending a spell slot
    or material components.

    **Prerequisite: 9th level**

    """

    name = "Ascendant Step"
    at_will_spells = (spells.Levitate,)


class BeastSpeech(Invocation):
    """You can cast speak with animals at will, without expending a spell slot."""

    name = "Beast Speech"
    at_will_spells = (spells.SpeakWithAnimals,)


class BeguilingInfluence(Invocation):
    """You gain proficiency in the Deception and Persuasion skills."""

    name = "Beguiling Influence"
    needs_implementation = True


class BewitchingWhispers(Invocation):
    """You can cast compulsion once using a warlock spell slot. You can't do so
    again until you finish a long rest

    **Prerequisite**: 7th Level
    """

    name = "Bewitching Whispers"


class BookOfAncientSecrets(Invocation):
    """You can now inscribe magical rituals in your Book of Shadows. Choose two
    1st-level spells that have the ritual tag from any class's spell list. The
    spells appear in the book and don't count against the number of spells you
    know. With your Book of Shadows in hand, you can cast the chosen spells as
    rituals. You can't cast the spells except as rituals, unless you've learned
    them by some other means. You can also cast a warlock spell you know as a
    ritual if it has the ritual tag.

    On your adventures, you can add other ritual spells to your Book o f
    Shadows. When you find such a spell, you can add it to the book if the
    spell's level is equal to or less than half your warlock level (rounded up)
    and if you can spare the time to transcribe the spell. For each level of
    the spell, the transcription process takes 2 hours and costs 50 gp for the
    rare inks needed to inscribe it

    """

    name = "Book of Ancient Secrets"


class ChainsOfCarceri(Invocation):
    """You can cast hold monster at will-targeting a celestial, fiend, or
    elemental-without expending a spell slot or material components. You must
    finish a long rest before you can use this invocation on the same creature
    again.

    **Prerequisites**: 15th level, Pact of the Chain Feature
    """

    name = "Chains of Carceri"


class DevilsSight(Invocation):
    """You can see normally in darkness, both magical and nonmagical, to a
    distance of 120 feet.

    """

    name = "Devil's Sight"


class DreadfulWord(Invocation):
    """You can cast confusion once using a warlock spell slot. You can't do so
    again until you finish a long rest.

    """

    name = "Dreadful Word"


class EldritchSight(Invocation):
    """You can cast detect magic at will, without expending a spell slot."""

    name = "Eldritch Sight"
    at_will_spells = (spells.DetectMagic,)


class EldritchSpear(Invocation):
    """When you cast eldritch blast, its range is 300 feet."""

    name = "Eldritch Spear"


class EyesOfTheRuneKeeper(Invocation):
    """
    You can read all writing.

    """

    name = "Eyes of the Rune Keeper"


class FiendishVigor(Invocation):
    """You can cast false life on yourself at will as a 1st-level spell, without
    expending a spell slot or material components.

    """

    name = "Fiendish Vigor"
    at_will_spells = (spells.FalseLife,)


class GazeOfTwoMinds(Invocation):
    """You can use your action to touch a willing humanoid and perceive through
    its senses until the end of your next turn. As long as the creature is on
    the same plane of existence as you, you can use your action on subsequent
    turns to maintain this connection, extending the duration until the end of
    your next turn. While perceiving through the other creature's senses, you
    benefit from any special senses possessed by that creature, and you are
    blinded and deafened to your own surroundings.

    """

    name = "Gaze of Two Minds"


class LifeDrinker(Invocation):
    """When you hit a creature with your pact weapon, the creature takes extra
    necrotic damage equal to your Charisma modifier (minimum 1).

    **Prerequisite**: 12th Level, Pact of the Blade
    """

    name = "Life Drinker"
    needs_implementation = True


class MaskOfManyFaces(Invocation):
    """You can cast disguise self at will, without expending a spell slot."""

    name = "Mask of Many Faces"
    at_will_spells = (spells.DisguiseSelf,)


class MasterOfMyriadForms(Invocation):
    """
    You can cast alter self at will, without expending a spell slot.

    **Prerequisite**: 15th Level
    """

    name = "Master of Myriad Forms"
    at_will_spells = (spells.AlterSelf,)


class MinionsOfChaos(Invocation):
    """You can cast conjure elemental once using a warlock spell slot. You can't
    do so again until you finish a long rest.

    **Prerequisite**: 9th Level
    """

    name = "Minions of Chaos"


class MireTheMind(Invocation):
    """You can cast slow once using a warlock spell slot. You can't do so again
    until you finish a long rest.

    """

    name = "Mire the Mind"


class MistyVisions(Invocation):
    """You can cast silent image at will, without expending a spell slot or
    material components.

    """

    name = "Misty Visions"
    at_will_spells = (spells.SilentImage,)


class OneWithShadows(Invocation):
    """When you are in an area of dim light or darkness, you can use your action
    to become invisible until you move or take an action or a reaction.

    **Prerequisite**: 5th Level
    """

    name = "One with Shadows"


class OtherworldlyLeap(Invocation):
    """You can cast jump on yourself at will, without expending a spell slot or
    material components.

    **Prerequisite**: 9th Level

    """

    name = "Otherworldly Leap"
    at_will_spells = (spells.Jump,)


class RepellingBlast(Invocation):
    """When you hit a creature with eldritch blast, you can push the creature up
    to 10 feet away from you in a straight line.

    """

    name = "Repelling Blast"


class SculptorOfFlesh(Invocation):
    """You can cast polymorph once using a warlock spell slot. You can't do so
    again until you finish a long rest.

    **Prerequisite**: 7th Level
    """

    name = "Sculptor of Flesh"


class SignOfIllOmen(Invocation):
    """You can cast bestow curse once using a warlock spell slot. You can't do so
    again until you finish a long rest.

    **Prerequisite**: 5th Level

    """

    name = "Sign of Ill Omen"


class ThiefOfFiveFates(Invocation):
    """You can cast bane once using a warlock spell slot. You can't do so again
    until you finish a long rest.

    """

    name = "Thief of Five Fates"


class ThirstingBlade(Invocation):
    """You can attack with your pact weapon twice, instead of once, whenever you
    take the Attack action on your turn.

    **Prerequisite**: 5th Level, Pact of the Blade
    """

    name = "Thirsting Blade"


class VisionsOfDistantRealms(Invocation):
    """
    You can cast arcane eye at will, without expending a spell slot

    **Prerequisite**: 15th level
    """

    name = "Visions of Distant Realms"
    at_will_spells = (spells.ArcaneEye,)


class VoiceOfTheChainMaster(Invocation):
    """You can communicate telepathically with your familiar and perceive through
    your familiar's senses as long as you are on the same plane of
    existence. Additionally, while perceiving through your familiar's senses,
    you can also speak through your familiar in your own voice, even if your
    familiar is normally incapable of speech.

    **Prerequisite**: Pact of the Chain

    """

    name = "Voice of the Chain Master"


class WhispersOfTheGrave(Invocation):
    """You can cast speak with dead at will, without expending a spell slot

    **Prerequsite**: 9th Level

    """

    name = "Whispers of the Grave"
    at_will_spells = (spells.SpeakWithDead,)


class WitchSight(Invocation):
    """You can see the true form of any shapechanger or creature concealed by
    illusion or transmutation magic while the creature is within 30 feet of you
    and within line of sight.

    """

    name = "Witch Sight"


# XGTE
class AspectOfTheMoon(Invocation):
    """You no longer need to sleep and can't be forced to sleep by any means. To
    gain the benefits of a long rest, you can spend all 8 hours doing light
    activity, such as reading your Book of Shadows and keeping watch.

    **Prerequisite**: Pact of the Tome
    """

    name = "Aspect of the Moon"


class CloakOfFlies(Invocation):
    """As a bonus action, you can surround yourselfwith a magical aura that looks
    like buzzing flies. The aura extends 5 feet from you in every direction,
    but not through total cover. It lasts until you're incapacitated or you
    dismiss it as a bonus action.

    The aura grants you advantage on Charisma
    (Intimidation) checks but disadvantage on all other Charisma checks. Any
    other creature that starts its turn in the aura takes poison damage equal
    to your Charisma modifier (minimum of O damage).

    Once you use this invocation, you can't use it again until you finish a
    short or long rest.

    **Prerequisite**: 5th level
    """

    name = "Cloak of Flies"


class EldritchSmite(Invocation):
    """Once per turn when you hit a creature with your pact weapon, you can expend
    a warlock spell slot to deal an extra 1d8 force damage to the target, plus
    another 1d8 per level of the spell slot, and you can knock the target
    prone if it is Huge or smaller.

    **Prerequisite**: 5th level, Pact of the Blade

    """

    name = "Eldritch Smite"


class GhostlyGaze(Invocation):
    """As an action, you gain the ability to see through solid objects to a range
    of 30 feet. Within that range, you have darkvision if you don't already
    have it. This special sight lasts for 1 minute or until your concentration
    ends (as if you were concentrating on a spell). During that time, you
    perceive objects as ghostly, transparent images.

    Once you use this invocation, you can't use it again until you finish a
    short or long rest.

    **Prerequisite**: 7th level
    """

    name = "Ghostly Gaze"


class GiftOfTheDepths(Invocation):
    """You can breathe underwater, and you gain a swimming speed equal to your
    walking speed. You can also cast water breathing once without expending a
    spell slot. You regain the ability to do so when you finish a long rest.

    **Prerequisite**: 5th level
    """

    name = "Gift of the Depths"


class GiftOfTheEverLivingOnes(Invocation):
    """Whenever you regain hit points while your familiar is within 100 feet
    ofyou, treat any dice rolled to determine the hit points you regain as
    having rolled their maximum value for you.

    **Prerequisite**: Pact of the Chain
    """

    name = "Gift of the Ever-Living Ones"


class GraspOfHadar(Invocation):
    """Once on each ofyour turns when you hit a creature with your eldritcli
    blast, you can move that creature in a straight line 10 feet closer to you

    """

    name = "Grasp of Hadar"


class ImprovedPactWeapon(Invocation):
    """You can use any weapon you summon with your Pact of the Blade feature as a
    spellcasting focus for your waru lock spells.

    In addition, the weapon gains a +1 bonus to its attack and damage rolls,
    unless it is a magic weapon that already has a bonus to those rolls.

    Finally, the weapon you conjure can be a shortbow, longbow, light crossbow,
    or heavy crossbow.

    **Prerequisite**: Pact of the Blade
    """

    name = "Improved Pact Weapon"

    def weapon_func(self, weapon: weapons.Weapon, **kwargs):
        """
        Add +1 to attack and damage if magic is not already magic
        """
        if (weapon.attack_bonus == 0) or (weapon.damage_bonus == 0):
            weapon.attack_bonus += 1
            weapon.damage_bonus += 1


class LanceOfLethargy(Invocation):
    """Once on each of your turns when you hit a creature with your eldritch
    blast, you can reduce that creature's speed by 10 feet until the end ofyour
    next turn.

    """

    name = "Lance of Lethargy"


class MaddeningHex(Invocation):
    """As a bonus action, you cause a psychic disturbance around the target cursed
    by your hex spell or by a warlock feature of yours, such as Hexblade's
    Curse or Sign of Ill Omen. When you do so, you deal psychic damage to the
    cursed target and each creature of your choice that you can see within 5
    feet of it. The psychic damage equals your Charisma modifier (minimum of 1
    damage). To use this invocation, you must be able to see the cursed
    target, and it must be within 30 feet ofyou.

    **Prerequisite**: 5th level
    """

    name = "Maddening Hex"


class RelentlessHex(Invocation):
    """Your curse creates a temporary bond between you and your target. As a bonus
    action, you can magically telerport up to 30 feet to an unoccupied space
    you can see within 5 feet of the target cursed by your hex spell or by a
    warlock feature ofyours, such as Hexblade's Curse or Sign of 111 Omen. To
    teleport in this way, you must be able to see the cursed target.

    """

    name = "Relentless Hex"


class ShroudOfShadow(Invocation):
    """You can cast invisibility at will, without expending a spell slot.

    **Prerequisite**: 15th Level

    """

    at_will_spells = (spells.Invisibility,)
    name = "Shroud of Shadow"


class TombOfLevistus(Invocation):
    """As a reaction when you take damage, you can entomb yourself in ice, which
    melts away at the end ofyour next turn. You gain 10 temporary hit points
    per warlock level, which take as much of the triggering damage as
    possible. Immediately after you take the damage, you gain vulnerability to
    fire damage, your speed is reduced to 0, and you are incapacitated. These
    effects, including any remaining temporary hit points, all end when the ice
    melts.

    Once you use this invocation, you can't use it again until you finish a
    short or long rest.

    **Prerequisite**: 5th Level

    """

    name = "Tomb of Levistus"


class TrickstersEscape(Invocation):
    """You can cast freedom ofmovement once on yourself without expending a spell
    slot. You regain the ability to do so when you finish a long rest

    """

    name = "Tricksters Escape"



"""
A collection of monsters from D&D 5e's SRD

This file was autogenerated by https://github.com/stravajiaxen/5e-srd-to-py
"""

from dungeonsheets.monsters.monsters import Monster
from dungeonsheets.stats import Ability


class Nalfeshnee(Monster):
    """
    Magic Resistance.
      The nalfeshnee has advantage on saving throws against spells and other
      magical effects.
    Multiattack.
      The nalfeshnee uses Horror Nimbus if it can. It then makes three
      attacks: one with its bite and two with its claws.
    Bite.
      Melee Weapon Attack: +10 to hit, reach 5 ft., one target. Hit: 32
      (5d10 + 5) piercing damage.
    Claw.
      Melee Weapon Attack: +10 to hit, reach 10 ft., one target. Hit: 15
      (3d6 + 5) slashing damage.
    Horror Nimbus.
      The nalfeshnee magically emits scintillating, multicolored light. Each
      creature within 15 feet of the nalfeshnee that can see the light must
      succeed on a DC 15 Wisdom saving throw or be frightened for 1 minute.
      A creature can repeat the saving throw at the end of each of its
      turns, ending the effect on itself on a success. If a creature's
      saving throw is successful or the effect ends for it, the creature is
      immune to the nalfeshnee's Horror Nimbus for the next 24 hours.
    Teleport.
      The nalfeshnee magically teleports, along with any equipment it is
      wearing or carrying, up to 120 feet to an unoccupied space it can see.
    Variant: Summon Demon.
      The demon chooses what to summon and attempts a magical summoning.
      
      A nalfeshnee has a 50 percent chance of summoning 1d4 vrocks, 1d3
      hezrous, 1d2 glabrezus, or one nalfeshnee.
      
      A summoned demon appears in an unoccupied space within 60 feet of its
      summoner, acts as an ally of its summoner, and can't summon other
      demons. It remains for 1 minute, until it or its summoner dies, or
      until its summoner dismisses it as an action.
    """
    name = "Nalfeshnee"
    description = "Large fiend, chaotic evil"
    challenge_rating = 13
    armor_class = 18
    skills = ""
    senses = "Truesight 120 ft., Passive Perception 11"
    languages = "Abyssal, telepathy 120 ft."
    strength = Ability(21)
    dexterity = Ability(10)
    constitution = Ability(22)
    intelligence = Ability(19)
    wisdom = Ability(12)
    charisma = Ability(15)
    speed = 20
    swim_speed = 0
    fly_speed = 30
    climb_speed = 0
    hp_max = 184
    hit_dice = "16d10"
    spells = []


class NightHag(Monster):
    """
    Innate Spellcasting.
      The hag's innate spellcasting ability is Charisma (spell save DC 14,
      +6 to hit with spell attacks). She can innately cast the following
      spells, requiring no material components:
      
  
      
      At will: detect magic, magic missile
      
      2/day each: plane shift (self only), ray of enfeeblement, sleep
    Magic Resistance.
      The hag has advantage on saving throws against spells and other
      magical effects.
    Night Hag Items.
      A night hag carries two very rare magic items that she must craft for
      herself If either object is lost, the night hag will go to great
      lengths to retrieve it, as creating a new tool takes time and effort.
      
      Heartstone: This lustrous black gem allows a night hag to become
      ethereal while it is in her possession. The touch of a heartstone also
      cures any disease. Crafting a heartstone takes 30 days.
      
      Soul Bag: When an evil humanoid dies as a result of a night hag's
      Nightmare Haunting, the hag catches the soul in this black sack made
      of stitched flesh. A soul bag can hold only one evil soul at a time,
      and only the night hag who crafted the bag can catch a soul with it.
      Crafting a soul bag takes 7 days and a humanoid sacrifice (whose flesh
      is used to make the bag).
    Claws (Hag Form Only).
      Melee Weapon Attack: +7 to hit, reach 5 ft., one target. Hit: 13 (2d8
      + 4) slashing damage.
    Change Shape.
      The hag magically polymorphs into a Small or Medium female humanoid,
      or back into her true form. Her statistics are the same in each form.
      Any equipment she is wearing or carrying isn't transformed. She
      reverts to her true form if she dies.
    Etherealness.
      The hag magically enters the Ethereal Plane from the Material Plane,
      or vice versa. To do so, the hag must have a heartstone in her
      possession.
    Nightmare Haunting.
      While on the Ethereal Plane, the hag magically touches a sleeping
      humanoid on the Material Plane. A protection from evil and good spell
      cast on the target prevents this contact, as does a magic circle. As
      long as the contact persists, the target has dreadful visions. If
      these visions last for at least 1 hour, the target gains no benefit
      from its rest, and its hit point maximum is reduced by 5 (1d10). If
      this effect reduces the target's hit point maximum to 0, the target
      dies, and if the target was evil, its soul is trapped in the hag's
      soul bag. The reduction to the target's hit point maximum lasts until
      removed by the greater restoration spell or similar magic.
    """
    name = "Night Hag"
    description = "Medium fiend, neutral evil"
    challenge_rating = 5
    armor_class = 17
    skills = "Deception +7, Insight +6, Perception +6, Stealth +6"
    senses = "Darkvision 120 ft., Passive Perception 16"
    languages = "Abyssal, Common, Infernal, Primordial"
    strength = Ability(18)
    dexterity = Ability(15)
    constitution = Ability(16)
    intelligence = Ability(16)
    wisdom = Ability(14)
    charisma = Ability(16)
    speed = 30
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 112
    hit_dice = "15d8"
    spells = []


class Nightmare(Monster):
    """
    Confer Fire Resistance.
      The nightmare can grant resistance to fire damage to anyone riding it.
    Illumination.
      The nightmare sheds bright light in a 10-foot radius and dim light for
      an additional 10 feet.
    Hooves.
      Melee Weapon Attack: +6 to hit, reach 5 ft., one target. Hit: 13 (2d8
      + 4) bludgeoning damage plus 7 (2d6) fire damage.
    Ethereal Stride.
      The nightmare and up to three willing creatures within 5 feet of it
      magically enter the Ethereal Plane from the Material Plane, or vice
      versa.
    """
    name = "Nightmare"
    description = "Large fiend, neutral evil"
    challenge_rating = 3
    armor_class = 13
    skills = ""
    senses = "Passive Perception 11"
    languages = "understands Abyssal, Common, and Infernal but can't speak"
    strength = Ability(18)
    dexterity = Ability(15)
    constitution = Ability(16)
    intelligence = Ability(10)
    wisdom = Ability(13)
    charisma = Ability(15)
    speed = 60
    swim_speed = 0
    fly_speed = 90
    climb_speed = 0
    hp_max = 68
    hit_dice = "8d10"
    spells = []


class Noble(Monster):
    """
    Rapier.
      Melee Weapon Attack: +3 to hit, reach 5 ft., one target. Hit: 5 (1d8 +
      1) piercing damage.
    """
    name = "Noble"
    description = "Medium humanoid, any alignment"
    challenge_rating = 0.125
    armor_class = 15
    skills = "Deception +5, Insight +4, Persuasion +5"
    senses = "Passive Perception 12"
    languages = "any two languages"
    strength = Ability(11)
    dexterity = Ability(12)
    constitution = Ability(11)
    intelligence = Ability(12)
    wisdom = Ability(14)
    charisma = Ability(16)
    speed = 30
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 9
    hit_dice = "2d8"
    spells = []


class Nothic(Monster):
    """Keen Sight
      The nothic has advantage on Wisdom (Perception) checks that rely
      on sight.
    Multiattack
      The nothic makes two claw attacks.
    Claw
      Melee Weapon Attack: +4 to hit, reach 5 ft., one target. Hit: 6
      (1d6 + 3) slashing damage.
    Rotting Gaze
      The nothic targets one creature it can see within 30 feet of
      it. The target must succeed on a DC 12 Constitution saving throw
      against this magic or take 10 (3d6) necrotic damage.
    Weird Insight
      The nothic targets one creature it can see within 30 feet of
      it. The target must contest its Charisma (Deception) check
      against the nothic’s Wisdom (Insight) check. If the nothic wins,
      it magically learns one fact or secret about the target. The
      target automatically wins if it is immune to being charmed.
    """
    name = "Nothic"
    description = "Medium aberration, neutral evil"
    challenge_rating = 2.0
    armor_class = 15
    skills = "Arcana +3, Insight +4, Perception +2, Stealth +5"
    senses = "Truesight 120ft., Passive Perception 12"
    languages = "Undercommon"
    strength = Ability(14)
    dexterity = Ability(16)
    constitution = Ability(16)
    intelligence = Ability(13)
    wisdom = Ability(10)
    charisma = Ability(8)
    speed = 30
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 45
    hit_dice = "6d8+18"
    spells = []



from dungeonsheets import spells
from dungeonsheets.features.features import Feature


# Cleric Features
class ChannelDivinity(Feature):
    """At 2nd level, you gain the ability to channel divine energy directly from
    your deity, using that energy to fuel magical effects. You start with two
    such effects: Turn Undead and an effect determined by your domain. Some
    domains grant you additional effects as you advance in levels, as noted in
    the domain description.

    When you use your Channel Divinity, you choose which effect to create. You
    must then finish a short or long rest to use your Channel Divinity again.

    Some Channel Divinity effects require saving throws.  When you use such an
    effect from this class, the DC equals your cleric spell save DC.

    Beginning at 6th level, you can use your Channel Divinity twice between
    rests, and beginning at 18th level you can use it three times between
    rests. When you finish a short or long rest, you regain your expended uses.

    """

    _name = "Channel Divinity"
    source = "Cleric"

    @property
    def name(self):
        level = self.owner.Cleric.level
        if level < 6:
            return "Channel Divinity (1x/SR)"
        elif level < 18:
            return "Channel Divinity (2x/SR)"
        else:
            return "Channel Divinity (3x/SR)"


class TurnUndead(Feature):
    """As an action, you present your holy symbol and speak a prayer censuring the
    undead. Each undead that can see or hear you within 30 feet of you must
    make a Wisdom saving throw. If the creature fails its saving throw, it is
    turned for 1 minute or until it takes any damage.

    A turned creature must spend its turns trying to move as far away from you
    as it can, and it can't willingly move to a space within 30 feet of you. It
    also can't take reactions. For its action, it can use only the Dash action
    or try to escape from an effect that prevents it from moving. If there's
    nowhere to move, the creature can use the Dodge action.

    """

    name = "Channel Divinity: Turn Undead"
    source = "Cleric"


class DestroyUndead(Feature):
    """Starting at 5th level, when an undead fails its saving throw against your
    Turn Undead feature, the creature is instantly destroyed if its challenge
    rating is at or below a certain threshold, as shown in the Destroy Undead
    table.

    """

    _name = "Destroy Undead"
    source = "Cleric"

    @property
    def name(self):
        level = self.owner.Cleric.level
        name = self._name + " (CR 1/2)"
        if level >= 8:
            name = self._name + " (CR 1)"
        if level >= 11:
            name = self._name + " (CR 2)"
        if level >= 14:
            name = self._name + " (CR 3)"
        if level >= 17:
            name = self._name + " (CR 4)"
        return name


class DivineIntervention(Feature):
    """Beginning at 10th level, you can call on your deity to intervene on your
    behalf when your need is great.

    Imploring your deity's aid requires you to use your action. Describe the
    assistance you seek, and roll percentile dice. If you roll a number equal
    to or lower than your cleric level, your deity intervenes. The DM chooses
    the nature of the intervention; the effect of any cleric spell or cleric
    domain spell would be appropriate.

    If your deity intervenes, you can't use this feature again for 7
    days. Otherwise, you can use it again after you finish a long rest.

    At 20th level, your call for intervention succeeds automatically, no roll
    required.

    """

    name = "Divine Intervention"
    source = "Cleric"


class DivineStrike(Feature):
    """
    Generic Divine Strike
    """

    _name = "Divine Strike"
    source = "Cleric"

    @property
    def name(self):
        level = self.owner.Cleric.level
        damage = " (1d8)"
        if level >= 14:
            damage = " (2d8)"
        return self._name + damage


# Knowledge Domain
class BlessingsOfKnowledge(Feature):
    """At 1st level, you learn two languages of your choice. You also become
    proficient in your choice o f two of the following skills: Arcana, History,
    Nature, or Religion. Your proficiency bonus is doubled for any ability
    check you make that uses either of those skills.

    """

    name = "Blessings of Knowledge"
    source = "Cleric (Knowledge Domain)"


class KnowledgeOfTheAncients(DivineIntervention):
    """Starting at 2nd level, you can use your Channel Divinity to tap into a
    divine well o f knowledge. As an action, you choose one skill or tool. For
    10 minutes, you have proficiency with the chosen skill or tool.

    """

    name = "Divine Intervention: Knowledge of the Ancients"
    source = "Cleric (Knowledge Domain)"


class ReadThoughts(DivineIntervention):
    """At 6th level, you can use your Channel Divinity to read a creature's
    thoughts. You can then use your access to the creature's mind to command
    it.  As an action, choose one creature that you can see within 60 feet of
    you. That creature must make a Wisdom saving throw. If the creature
    succeeds on the saving throw, you can't use this feature on it again until
    you finish a long rest.

    If the creature fails its save, you can read its surface thoughts (those
    foremost in its mind, reflecting its current emotions and what it is
    actively thinking about) when it is within 60 feet of you. This effect
    lasts for 1 minute. During that time, you can use your action to end this
    effect and cast the suggestion spell on the creature without expending a
    spell slot. The target automatically fails its saving throw against the
    spell.

    """

    name = "Divine Intervention: Read Thoughts"
    source = "Cleric (Knowledge Domain)"


class PotentSpellcasting(Feature):
    """Starting at 8th level, you add your W isdom modifier to the damage you deal
    with any cleric cantrip.

    """

    name = "Potent Spellcasting"
    source = "Cleric"


class VisionsOfThePast(Feature):
    """Starting at 17th level, you can call up visions of the past that relate to
    an object you hold or your immediate surroundings. You spend at least 1
    minute in meditation and prayer, then receive dreamlike, shadowy glimpses
    of recent events. You can meditate in this way for a number of minutes
    equal to your Wisdom score and must maintain concentration during that
    time, as if you were casting a spell. Once you use this feature, you can't
    use it again until you finish a short or long rest.

    **Object Reading**: Holding an object as you meditate, you can see visions
    of the object's previous owner. After meditating for 1 minute, you learn
    how the owner acquired and lost the object, as well as the most recent
    significant event involving the object and that owner. If the object was
    owned by another creature in the recent past (within a number of days equal
    to your Wisdom score), you can spend 1 additional minute for each owner to
    learn the same information about that creature.

    **Area Reading**: As you meditate, you see visions of recent events in your
    immediate vicinity (a room, street, tunnel, clearing, or the like, up to a
    50-foot cube), going back a number of days equal to your Wisdom score. For
    each minute you meditate, you learn about one significant event, beginning
    with the most recent. Significant events typically involve powerful
    emotions, such as battles and betrayals, marriages and murders, births and
    funerals. However, they might also include more mundane events that are
    nevertheless important in your current situation.

    """

    name = "Visions of the Past"
    source = "Cleric (Knowledge Domain)"


# Life Domain
class DiscipleOfLife(Feature):
    """Also starting at 1st level, your healing spells are more
    effective. Whenever you use a spell of 1st level or higher to restore hit
    points to a creature, the creature regains additional hit points equal to 2
    + the spell's level

    """

    name = "Disciple of Life"
    source = "Cleric (Life Domain)"


class PreserveLife(ChannelDivinity):
    """Starting at 2nd level, you can use your Channel Divinity to heal the badly
    injured. As an action, you present your holy symbol and evoke healing
    energy that can restore a number of hit points equal to five times your
    cleric level. Choose any creatures within 30 feet of you, and divide those
    hit points among them. This feature can restore a creature to no more than
    half of its hit point maximum. You can't use this feature on an undead or a
    construct.

    """

    name = "Channel Divinity: Preserve Life"
    source = "Cleric (Life Domain)"


class BlessedHealer(Feature):
    """Beginning at 6th level, the healing spells you cast on others heal you as
    well. When you cast a spell of 1st level or higher that restores hit points
    to a creature other than you, you regain hit points equal to 2 + the
    spell's level.

    """

    name = "Blessed Healer"
    source = "Cleric (Life Domain)"


class DivineStrikeLife(DivineStrike):
    """At 8th level, you gain the ability to infuse your weapon strikes with
    divine energy. Once on each o f your turns when you hit a creature with a
    weapon attack, you can cause the attack to deal an extra 1d8 radiant damage
    to the target. When you reach 14th level, the extra damage increases to
    2d8.

    """

    source = "Cleric (Life Domain)"


class SupremeHealing(Feature):
    """Starting at 17th level, when you would normally roll one or more dice to
    restore hit points with a spell, you instead use the highest number
    possible for each die. For example, instead of restoring 2d6 hit points to
    a creature, you restore 12.

    """

    name = "Supreme Healing"
    source = "Cleric (Life Domain)"


# Light Domain
class WardingFlare(Feature):
    """At 1st level, you can interpose divine light between yourself and an
    attacking enemy. When you are attacked by a creature within 30 feet of you
    that you can see, you can use your reaction to impose disadvantage on the
    attack roll, causing light to flare before the attacker before it hits or
    misses. An attacker that can't be blinded is immune to this feature. You
    can use this feature a number of times equal to your Wisdom modifier (a
    minimum of once). You regain all expended uses when you finish a long rest

    """

    _name = "Warding Flare"
    source = "Cleric (Light Domain)"

    @property
    def name(self):
        times = max(1, self.owner.wisdom.modifier)
        return self._name + " ({:d}x/LR)".format(times)


class RadianceOfTheDawn(ChannelDivinity):
    """Starting at 2nd level, you can use your Channel Divinity to harness
    sunlight, banishing darkness and dealing radiant damage to your foes. As an
    action, you present your holy symbol, and any magical darkness within 30
    feet of you is dispelled.

    Additionally, each hostile creature within 30 feet of you must make a
    Constitution saving throw. A creature takes radiant damage equal to 2d10 +
    your cleric level on a failed saving throw, and half as much damage on a
    successful one. A creature that has total cover from you is not affected.

    """

    name = "Channel Divinity: Radiance of the Dawn"
    source = "Cleric (Light Domain)"


class ImprovedFlare(Feature):
    """Starting at 6th level, you can also use your Warding Flare feature when a
    creature that you can see within 30 feet o f you attacks a creature other
    than you.

    """

    name = "Improved Flare"
    source = "Cleric (Light Domain)"


class CoronaOfLight(Feature):
    """Starting at 17th level, you can use your action to activate an aura of
    sunlight that lasts for 1 minute or until you dismiss it using another
    action. You emit bright light in a 60-foot radius and dim light 30 feet
    beyond that. Your enemies in the bright light have disadvantage on saving
    throws against any spell that deals fire or radiant damage

    """

    name = "Corona of Light"
    source = "Cleric (Light Domain)"


# Nature Domain
class AcolyteOfNature(Feature):
    """At 1st level, you learn one druid cantrip of your choice. You also gain
    proficiency in one of the following skills of your choice: Animal Handling,
    Nature, or Survival.

    """

    name = "Acolyte of Nature"
    source = "Cleric (Nature Domain)"


class CharmAnimalsAndPlants(ChannelDivinity):
    """Starting at 2nd level, you can use your Channel Divinity to charm animals
    and plants. As an action, you present your holy symbol and invoke the name
    of your deity. Each beast or plant creature that can see you within 30 feet
    of you must make a Wisdom saving throw.

    If the creature fails its saving throw, it is charmed by you for 1 minute
    or until it takes damage. While it is charmed by you, it is friendly to you
    and other creatures you designate.

    """

    name = "Channel Divinity: Charm Animals and Plants"
    source = "Cleric (Nature Domain)"


class DampenElements(Feature):
    """Starting at 6th level, when you or a creature within 30 feet of you takes
    acid, cold, fire, lightning, or thunder damage, you can use your reaction
    to grant resistance to the creature against that instance of the damage.

    """

    name = "Dampen Elements"
    source = "Cleric (Nature Domain)"


class DivineStrikeNature(DivineStrike):
    """At 8th level, you gain the ability to infuse your weapon strikes with
    divine energy. Once on each of your turns when you hit a creature with a
    weapon attack, you can cause the attack to deal an extra 1d8 cold, fire, or
    lightning damage (your choice) to the target. When you reach 14th level,
    the extra damage increases to 2d8.

    """

    source = "Cleric (Nature Domain)"


class MasterOfNature(Feature):
    """At 17th level, you gain the ability to command animals and plant
    creatures. While creatures are charmed by your Charm Animals and Plants
    feature, you can take a bonus action on your turn to verbally command what
    each of those creatures will do on its next turn.

    """

    name = "Master of Nature"
    source = "Cleric (Nature Domain)"


# Tempest Domain
class WrathOfTheStorm(Feature):
    """Also at 1st level, you can thunderously rebuke attackers. When a creature
    within 5 feet of you that you can see hits you with an attack, you can use
    your reaction to cause the creature to make a Dexterity saving throw. The
    creature takes 2d8 lightning or thunder damage (your choice) on a failed
    saving throw, and half as much damage on a successful one.

    You can use this feature a number of times equal to your Wisdom modifier (a
    minimum of once). You regain all expended uses when you finish a long rest.

    """

    _name = "Wrath of the Storm"
    source = "Cleric (Tempest Domain)"

    @property
    def name(self):
        num_uses = max(1, self.owner.wisdom.modifier)
        return self._name + " ({:d}x/LR)".format(num_uses)


class DestructiveWrath(ChannelDivinity):
    """Starting at 2nd level, you can use your Channel Divinity to wield the power
    of the storm with unchecked ferocity.

    When you roll lightning or thunder damage, you can use your Channel
    Divinity to deal maximum damage, instead of rolling.

    """

    name = "Channel Divinity: Destructive Wrath"
    source = "Cleric (Tempest Domain)"


class ThunderboltStrike(Feature):
    """At 6th level, when you deal lightning damage to a Large or smaller
    creature, you can also push it up to 10 feet away from you.

    """

    name = "Thunderbolt Strike"
    source = "Cleric (Tempest Domain)"


class DivineStrikeTempest(DivineStrike):
    """At 8th level, you gain the ability to infuse your weapon strikes with
    divine energy. Once on each of your turns when you hit a creature with a
    weapon attack, you can cause the attack to deal an extra 1d8 thunder damage
    to the target. When you reach 14th level, the extra damage increases to
    2d8.

    """

    source = "Cleric (Tempest Domain)"


class Stormborn(Feature):
    """At 17th level, you have a flying speed equal to your current walking speed
    whenever you are not underground or indoors.

    """

    name = "Stormborn"
    source = "Cleric (Tempest Domain)"


# Trickery Domain
class BlessingOfTheTrickster(Feature):
    """Starting when you choose this domain at 1st level, you can use your action
    to touch a willing creature other than yourself to give it advantage on
    Dexterity (Stealth) checks. This blessing lasts for 1 hour or until you use
    this feature again

    """

    name = "Blessing of the Trickster"
    source = "Cleric (Trickery Domain)"


class InvokeDuplicity(ChannelDivinity):
    """Starting at 2nd level, you can use your Channel Divinity to create an
    illusory duplicate o f yourself. As an action, you create a perfect
    illusion of yourself that lasts for 1 minute, or until you lose your
    concentration (as if you were concentrating on a spell). The illusion
    appears in an unoccupied space that you can see within 30 feet of you. As a
    bonus action on your turn, you can move the illusion up to 30 feet to a
    space you can see, but it must remain within 120 feet of you.

    For the duration, you can cast spells as though you were in the illusion's
    space, but you must use your own senses. Additionally, when both you and
    your illusion are within 5 feet of a creature that can see the illusion,
    you have advantage on attack rolls against that creature, given how
    distracting the illusion is to the target.

    """

    name = "Channel Divinity: Invoke Duplicity"
    source = "Cleric (Trickery Domain)"


class CloakOfShadows(ChannelDivinity):
    """Starting at 6th level, you can use your Channel Divinity to vanish. As an
    action, you become invisible until the end of your next turn. You become
    visible if you attack or cast a spell.

    """

    name = "Channel Divinity: Cloak of Shadows"
    source = "Cleric (Trickery Domain)"


class DivineStrikeTrickery(DivineStrike):
    """At 8th level, you gain the ability to infuse your weapon strikes with
    poison-a gift from your deity. Once on each of your turns when you hit a
    creature with a weapon attack, you can cause the attack to deal an extra
    1d8 poison damage to the target. When you reach 14th level, the extra
    damage increases to 2d8.

    """

    source = "Cleric (Trickery Domain)"


class ImprovedDuplicity(Feature):
    """At 17th level, you can create up to four duplicates of yourself, instead of
    one, when you use Invoke Duplicity. As a bonus action on your turn, you can
    move any number of them up to 30 feet, to a maximum range of 120 feet.

    """

    name = "Improved Duplicity"
    source = "Cleric (Trickery Domain)"


# War Domain
class WarPriest(Feature):
    """From 1st level, your god delivers bolts of inspiration to you while you are
    engaged in battle. When you use the Attack action, you can make one weapon
    attack as a bonus action. You can use this feature a number of times equal
    to your Wisdom modifier (a minimum of once). You regain all expended uses
    when you finish a long rest

    """

    _name = "War Priest"
    source = "Cleric (War Domain)"

    @property
    def name(self):
        num = max(1, self.owner.wisdom.modifier)
        return self._name + " ({:d}x/LR)".format(num)


class GuidedStrike(ChannelDivinity):
    """Starting at 2nd level, you can use your Channel Divinity to strike with
    supernatural accuracy. When you make an attack roll, you can use your
    Channel Divinity to gain a +10 bonus to the roll. You make this choice
    after you see the roll, but before the DM says whether the attack hits or
    misses

    """

    name = "Channel Divinity: Guided Strike"
    source = "Cleric (War Domain)"


class WarGodsBlessing(ChannelDivinity):
    """At 6th level, when a creature within 30 feet of you makes an attack roll,
    you can use your reaction to grant that creature a +10 bonus to the roll,
    using your Channel Divinity. You make this choice after you see the roll,
    but before the DM says whether the attack hits or misses.

    """

    name = "Channel Divinity: War Gods Blessing"
    source = "Cleric (War Domain)"


class DivineStrikeWar(DivineStrike):
    """At 8th level, you gain the ability to infuse your weapon strikes with
    divine energy. Once on each of your turns when you hit a creature with a
    weapon attack, you can cause the attack to deal an extra 1d8 damage of the
    same type dealt by the weapon to the target. When you reach 14th level, the
    extra damage increases to 2d8.
    """

    source = "Cleric (War Domain)"


class AvatarOfBattle(Feature):
    """At 17th level, you gain resistance to bludgeoning, piercing, and slashing
    damage from nonmagical weapons.

    """

    name = "Avatar of Battle"
    source = "Cleric (War Domain)"


# Arcana Domain
class ArcaneInitiate(Feature):
    """When you choose this domain at 1st level, you gain proficiency in the
    Arcana skill, and you gain two cantrips of your choice from the wizard
    spell list. For you, these cantrips count as cleric cantrips

    """

    name = "Arcane Initiate"
    source = "Cleric (Arcana Domain)"


class ArcaneAbjuration(ChannelDivinity):
    """Starting at 2nd level, you can use your Channel Divinity to abjure
    otherworldly creatures. As an action, you present your holy symbol, and one
    celestial, elemental, fey, or fiend of your choice that is within 30 feet
    of you must make a Wisdom saving throw, provided that the creature can see
    or hear you. If the creature fails its saving throw, it is turned for 1
    minute or until it takes any damage.

    A turned creature must spend its turns trying to move as far away from you
    as it can, and it can't willingly end its move in a space within 30 feet of
    you. It also can't take reactions. For its action, it can only use the Dash
    action or try to escape from an effect that prevents it from moving. If
    there's nowhere to move, the creature can use the Dodge action.

    After you reach 5th level, when a creature fails its saving throw against
    your Arcane Abjuration feature, the creature is banished for 1 minute (as
    in the banishment spell, no concentration required) if it isn't on its
    plane of origin and its challenge rating is at or below a certain
    threshold, as shown on the Arcane Banishment table.

    5th level : CR 1/2
    8th level : CR 1
    11th level : CR 2
    14th level : CR 3
    17th level : CR 4

    """

    name = "Channel Divinity: Arcane Abjuration"
    source = "Cleric (Arcana Domain)"


class SpellBreaker(Feature):
    """Starting at 6th level, when you restore hit points to an ally with a spell
    of 1st level or higher, you can also end one spell of your choice on that
    creature. The level of the spell you end must be equal to or lower than the
    level of the spell slot you use to cast the healing spell

    """

    name = "Spell Breaker"
    source = "Cleric (Arcana Domain)"


class ArcaneMastery(Feature):
    """At 17th level, you choose four spells from the wizard spell list, one from
    each of the following levels: 6th, 7th, 8th, and 9th. You add them to your
    list of domain spells. Like your other domain spells, they are always
    prepared and count as cleric spells for you.

    """

    name = "Arcane Mastery"
    source = "Cleric (Arcana Domain)"


# Forge Domain
class BlessingOfTheForge(Feature):
    """At 1st level, you gain the ability to imbue magic into a weapon or
    armor. At the end of a long rest, you can touch one nonmagical object that
    is a suit of armor or a simple or martial weapon. Until the end of your
    next long rest or until you die, the object becomes a magic item, granting
    a +1 bonus to AC if it's armor or a +1 bo- nus to attack and damage rolls
    if it's a weapon. Once you use this feature, you can't use it again until
    you finish a long rest

    """

    name = "Blessing of the Forge"
    source = "Cleric (Forge Domain)"


class ArtisansBlessing(Feature):
    """Starting at 2nd level, you can use your Channel Divinity to create simple
    items. You conduct an hour-long ritual that crafts a nonmagi- cal item that
    must include some metal: a simple or martial weapon, a suit of armor, ten
    pieces of ammunition, a set of tools, or another metal Object (see chapter
    5, "Equipment," in the Player's Handbook for examples of these items). The
    creation is completed at the end of the hour, coalescing in an unoccupied
    space of your choice on a surface within 5 feet of you. The thing you
    create can be something that is worth no more than 100 gp.

    As part of this ritual, you must lay out metal, which can include coins,
    with a value equal to the creation. The metal irretrievably coalesces and
    transforms into the creation at the ritual's end, magically forming even
    nonmetal parts of the creation. The ritual can create a duplicate of a
    nonmagical item that contains metal, such as a key, if you possess the
    original during the ritual.

    """

    name = "Channel Divinity: Artisans Blessing"
    source = "Cleric (Forge Domain)"


class SoulOfTheForge(Feature):
    """Starting at 6th level, your mastery of the forge grants you special
    abilities:

    • You gain resistance to fire damage.

    • While wearing heavy armor, you gain a +1 bonus to AC.

    """

    name = "Soul of the Forge"
    source = "Cleric (Forge Domain)"


class DivineStrikeForge(DivineStrike):
    """At 8th level, you gain the ability to infuse your weapon strikes
    with the fiery power of the forge. Once on each ofyour turns when
    you hit a creature with a weapon attack, you can cause the attack
    to deal an extra 1d8 fire damage to the target. When you reach
    14th level, the extra damage increases to 2d8

    """

    source = "Cleric (Forge Domain)"


class SaintOfForgeAndFire(Feature):
    """At 17th level, your blessed affinity with fire and metal becomes
    more powerful:

    - You gain immunity to fire damage.
    - While wearing heavy armor, you have resistance to bludgeoning,
      piercing, and slashing damage from non-magical attacks

    """

    name = "Saint of Forge and Fire"
    source = "Cleric (Forge Domain)"


# Grave Domain
class CircleOfMortality(Feature):
    """At 1st level, you gain the ability to manipulate the line between life and
    death. When you would normally roll one or more dice to restore hit points
    with a spell to a creature at 0 hit points, you instead use the highest
    number possible for each die. In addition, you learn the spare the dying
    cantrip, which doesn't count against the number of cleric cantrips you
    know. For you, it has a range of 30 feet, and you can cast it as a bonus
    action

    """

    spells_known = (spells.SpareTheDying,)
    spells_prepared = (spells.SpareTheDying,)
    name = "Circle of Mortality"
    source = "Cleric (Grave Domain)"


class EyesOfTheGrave(Feature):
    """At lst level, you gain the ability to occasionally sense the presence of the
    undead, whose existence is an insult to the natural cycle of life. As an
    action, you can open your awareness to magically detect undead. Until the
    end ofyour next turn, you know the location of any undead within 60 feet of
    you that isn't behind total cover and that isn't protected from divination
    magic.

    This sense doesn't tell you anything about a creature's capabilities or
    identity. You can use this feature a number of times equal to your Wisdom
    modifier (minimum Of once). You regain all expended uses when you finish a
    long rest

    """

    _name = "Eyes of the Grave"
    source = "Cleric (Grave Domain)"

    @property
    def name(self):
        num = max(1, self.owner.wisdom.modifier)
        return self._name + " ({:d}x/LR)".format(num)


class PathToTheGrave(ChannelDivinity):
    """Starting at 2nd level, you can use your Channel Divinity to mark another
    creature's life force for termination. As an action, you choose one
    creature you can see within 30 feet of you, cursing it until the end Of
    your next turn. The next time you or an ally Ofyours hits the cursed
    creature with an attack, the creature has vulnerability tO all of that
    attack's damage, and then the curse ends

    """

    name = "Channel Divinity: Path to the Grave"
    source = "Cleric (Grave Domain)"


class SentinelAtDeathsDoor(Feature):
    """At 6th level, you gain the ability to impede death's progress. As a
    reaction when you or a creature you can see within 30 feet of you suffers a
    critical hit, you can turn that hit into a normal hit. Any effects
    triggered by a critical hit are canceled. You can use this feature a
    number of times equal to your Wisdom modifier (minimum of once). You regain
    all expended uses when you finish a long rest.

    """

    _name = "Sentinel at Death's Door"
    source = "Cleric (Grave Domain)"

    @property
    def name(self):
        num = max(1, self.owner.wisdom.modifier)
        return self._name + " ({:d}x/LR)".format(num)


class KeeperOfSouls(Feature):
    """Starting at 17th level. you can seize a trace of vitality from a parting
    soul and use it to heal the living. When an enemy you can see dies within
    60 feet of you, you or one creature of your choice that is within 60 feet
    of you regains hit points equal to the enemy's number of Hit Dice. You can
    use this feature only if you aren't incapacitated. Once you use it, you
    can't do so again until the start ofyour next turn.

    """

    name = "Keeper of Souls"
    source = "Cleric (Grave Domain)"


class Reaper(Feature):
    """At 1st level, you learn one necromancy cantrip of your choice from any
    spell list. When you cast a necromancy cantrip that normally targets only
    one creature, the spell can instead target two creatures within range and
    within 5 feet of each other.

    """

    name = "Reaper"
    source = "Cleric (Death Domain)"


class TouchOfDeathCleric(Feature):
    """Starting at 2nd level, you can use Channel Divinity to destroy another
    creature's life force by touch. When you hit a creature with a melee
    attack, you can use Channel Divinity to deal extra necrotic damage to
    the target. The damage equals 5 + twice your cleric level.

    """

    name = "Channel Divinity: Touch of Death"
    source = "Cleric (Death Domain)"


class InescapableDestruction(Feature):
    """Starting at 6th level, your ability to channel negative energy becomes
    more potent. Necrotic damage dealt by your cleric spells and Channel
    Divinity options ignores resistance to necrotic damage

    """

    name = "Inescapable Destruction"
    source = "Cleric (Death Domain)"


class DivineStrikeDeath(DivineStrike):
    """At 8th level, you gain the ability to infuse your weapon strikes with
    necrotic energy. Once on each of your turns when you hit a creature with
    a weapon attack, you can cause the attack to deal an a 1d8 necrotic
    damage to the target. When you reach 14th level, the extra damage
    increases to 2d8.

    """

    source = "Cleric (Death Domain)"


class ImprovedReaper(Feature):
    """Starting at 17th level, when you cast a necromancy spell of 1st through
    5th level that targets only one creature, the spell can instead target two
    creatures within range and within 5 feet of each other. If the spell
    consumes its material components, you must provide them for each target.

    """

    name = "Improved Reaper"
    source = "Cleric (Death Domain)"



"""
A collection of monsters from D&D 5e's SRD

This file was autogenerated by https://github.com/stravajiaxen/5e-srd-to-py
"""

from dungeonsheets.monsters.monsters import Monster
from dungeonsheets.stats import Ability


class Panther(Monster):
    """
    Keen Smell.
      The panther has advantage on Wisdom (Perception) checks that rely on
      smell.
    Pounce.
      If the panther moves at least 20 ft. straight toward a creature and
      then hits it with a claw attack on the same turn, that target must
      succeed on a DC 12 Strength saving throw or be knocked prone. If the
      target is prone, the panther can make one bite attack against it as a
      bonus action.
    Bite.
      Melee Weapon Attack: +4 to hit, reach 5 ft., one target. Hit: 5 (1d6 +
      2) piercing damage.
    Claw.
      Melee Weapon Attack: +4 to hit, reach 5 ft., one target. Hit: 4 (1d4 +
      2) slashing damage.
    """
    name = "Panther"
    description = "Medium beast, unaligned"
    challenge_rating = 0.25
    armor_class = 12
    skills = "Perception +4, Stealth +6"
    senses = "Passive Perception 14"
    languages = ""
    strength = Ability(14)
    dexterity = Ability(15)
    constitution = Ability(10)
    intelligence = Ability(3)
    wisdom = Ability(14)
    charisma = Ability(7)
    speed = 50
    swim_speed = 0
    fly_speed = 0
    climb_speed = 40
    hp_max = 13
    hit_dice = "3d8"
    spells = []


class Pegasus(Monster):
    """
    Hooves.
      Melee Weapon Attack: +6 to hit, reach 5 ft., one target. Hit: 11 (2d6
      + 4) bludgeoning damage.
    """
    name = "Pegasus"
    description = "Large celestial, chaotic good"
    challenge_rating = 2
    armor_class = 12
    skills = "Perception +6"
    senses = "Passive Perception 16"
    languages = "understands Celestial, Common, Elvish, and Sylvan but can't speak"
    strength = Ability(18)
    dexterity = Ability(15)
    constitution = Ability(16)
    intelligence = Ability(10)
    wisdom = Ability(15)
    charisma = Ability(13)
    speed = 60
    swim_speed = 0
    fly_speed = 90
    climb_speed = 0
    hp_max = 59
    hit_dice = "7d10"
    spells = []


class PhaseSpider(Monster):
    """
    Ethereal Jaunt.
      As a bonus action, the spider can magically shift from the Material
      Plane to the Ethereal Plane, or vice versa.
    Spider Climb.
      The spider can climb difficult surfaces, including upside down on
      ceilings, without needing to make an ability check.
    Web Walker.
      The spider ignores movement restrictions caused by webbing.
    Bite.
      Melee Weapon Attack: +4 to hit, reach 5 ft., one creature. Hit: 7
      (1d10 + 2) piercing damage, and the target must make a DC 11
      Constitution saving throw, taking 18 (4d8) poison damage on a failed
      save, or half as much damage on a successful one. If the poison damage
      reduces the target to 0 hit points, the target is stable but poisoned
      for 1 hour, even after regaining hit points, and is paralyzed while
      poisoned in this way.
    """
    name = "Phase Spider"
    description = "Large monstrosity, unaligned"
    challenge_rating = 3
    armor_class = 13
    skills = "Stealth +6"
    senses = "Darkvision 60 ft., Passive Perception 10"
    languages = ""
    strength = Ability(15)
    dexterity = Ability(15)
    constitution = Ability(12)
    intelligence = Ability(6)
    wisdom = Ability(10)
    charisma = Ability(6)
    speed = 30
    swim_speed = 0
    fly_speed = 0
    climb_speed = 30
    hp_max = 32
    hit_dice = "5d10"
    spells = []


class PitFiend(Monster):
    """
    Fear Aura.
      Any creature hostile to the pit fiend that starts its turn within 20
      feet of the pit fiend must make a DC 21 Wisdom saving throw, unless
      the pit fiend is incapacitated. On a failed save, the creature is
      frightened until the start of its next turn. If a creature's saving
      throw is successful, the creature is immune to the pit fiend's Fear
      Aura for the next 24 hours.
    Magic Resistance.
      The pit fiend has advantage on saving throws against spells and other
      magical effects.
    Magic Weapons.
      The pit fiend's weapon attacks are magical.
    Innate Spellcasting.
      The pit fiend's spellcasting ability is Charisma (spell save DC 21).
      The pit fiend can innately cast the following spells, requiring no
      material components:
      
      At will: detect magic, fireball
      
      3/day each: hold monster, wall of fire
    Multiattack.
      The pit fiend makes four attacks: one with its bite, one with its
      claw, one with its mace, and one with its tail.
    Bite.
      Melee Weapon Attack: +14 to hit, reach 5 ft., one target. Hit: 22 (4d6
      + 8) piercing damage. The target must succeed on a DC 21 Constitution
      saving throw or become poisoned. While poisoned in this way, the
      target can't regain hit points, and it takes 21 (6d6) poison damage at
      the start of each of its turns. The poisoned target can repeat the
      saving throw at the end of each of its turns, ending the effect on
      itself on a success.
    Claw.
      Melee Weapon Attack: +14 to hit, reach 10 ft. , one target. Hit: 17
      (2d8 + 8) slashing damage.
    Mace.
      Melee Weapon Attack: +14 to hit, reach 10ft., one target. Hit: 15 (2d6
      + 8) bludgeoning damage plus 21 (6d6) fire damage.
    Tail.
      Melee Weapon Attack: +14 to hit, reach 10ft., one target. Hit: 24
      (3d1O + 8) bludgeoning damage.
    """
    name = "Pit Fiend"
    description = "Large fiend, lawful evil"
    challenge_rating = 20
    armor_class = 19
    skills = ""
    senses = "Truesight 120 ft., Passive Perception 14"
    languages = "Infernal, telepathy 120 ft."
    strength = Ability(26)
    dexterity = Ability(14)
    constitution = Ability(24)
    intelligence = Ability(22)
    wisdom = Ability(18)
    charisma = Ability(24)
    speed = 30
    swim_speed = 0
    fly_speed = 60
    climb_speed = 0
    hp_max = 300
    hit_dice = "24d10"
    spells = []


class Planetar(Monster):
    """
    Angelic Weapons.
      The planetar's weapon attacks are magical. When the planetar hits with
      any weapon, the weapon deals an extra 5d8 radiant damage (included in
      the attack).
    Divine Awareness.
      The planetar knows if it hears a lie.
    Innate Spellcasting.
      The planetar's spellcasting ability is Charisma (spell save DC 20).
      The planetar can innately cast the following spells, requiring no
      material components:
      
      At will: detect evil and good, invisibility (self only)
      
      3/day each: blade barrier, dispel evil and good, flame strike, raise
      dead
      
      1/day each: commune, control weather, insect plague
    Magic Resistance.
      The planetar has advantage on saving throws against spells and other
      magical effects.
    Multiattack.
      The planetar makes two melee attacks.
    Greatsword.
      Melee Weapon Attack: +12 to hit, reach 5 ft., one target. Hit: 21 (4d6
      + 7) slashing damage plus 22 (5d8) radiant damage.
    Healing Touch.
      The planetar touches another creature. The target magically regains 30
      (6d8 + 3) hit points and is freed from any curse, disease, poison,
      blindness, or deafness.
    """
    name = "Planetar"
    description = "Large celestial, lawful good"
    challenge_rating = 16
    armor_class = 19
    skills = "Perception +11"
    senses = "Truesight 120 ft., Passive Perception 21"
    languages = "all, telepathy 120 ft."
    strength = Ability(24)
    dexterity = Ability(20)
    constitution = Ability(24)
    intelligence = Ability(19)
    wisdom = Ability(22)
    charisma = Ability(25)
    speed = 40
    swim_speed = 0
    fly_speed = 120
    climb_speed = 0
    hp_max = 200
    hit_dice = "16d10"
    spells = []


class Plesiosaurus(Monster):
    """
    Hold Breath.
      The plesiosaurus can hold its breath for 1 hour.
    Bite.
      Melee Weapon Attack: +6 to hit, reach 10 ft., one target. Hit: 14 (3d6
      + 4) piercing damage.
    """
    name = "Plesiosaurus"
    description = "Large beast, unaligned"
    challenge_rating = 2
    armor_class = 13
    skills = "Perception +3, Stealth +4"
    senses = "Passive Perception 13"
    languages = ""
    strength = Ability(18)
    dexterity = Ability(15)
    constitution = Ability(16)
    intelligence = Ability(2)
    wisdom = Ability(12)
    charisma = Ability(5)
    speed = 20
    swim_speed = 40
    fly_speed = 0
    climb_speed = 0
    hp_max = 68
    hit_dice = "8d10"
    spells = []


class PoisonousSnake(Monster):
    """
    Bite.
      Melee Weapon Attack: +5 to hit, reach 5 ft., one target. Hit: 1
      piercing damage, and the target must make a DC 10 Constitution saving
      throw, taking 5 (2d4) poison damage on a failed save, or half as much
      damage on a successful one.
    """
    name = "Poisonous Snake"
    description = "Tiny beast, unaligned"
    challenge_rating = 0.125
    armor_class = 13
    skills = ""
    senses = "Blindsight 10 ft., Passive Perception 10"
    languages = ""
    strength = Ability(2)
    dexterity = Ability(16)
    constitution = Ability(11)
    intelligence = Ability(1)
    wisdom = Ability(10)
    charisma = Ability(3)
    speed = 30
    swim_speed = 30
    fly_speed = 0
    climb_speed = 0
    hp_max = 2
    hit_dice = "1d4"
    spells = []


class PolarBear(Monster):
    """
    Keen Smell.
      The bear has advantage on Wisdom (Perception) checks that rely on
      smell.
    Multiattack.
      The bear makes two attacks: one with its bite and one with its claws.
    Bite.
      Melee Weapon Attack: +7 to hit, reach 5 ft., one target. Hit: 9 (1d8 +
      5) piercing damage.
    Claws.
      Melee Weapon Attack: +7 to hit, reach 5 ft., one target. Hit: 12 (2d6
      + 5) slashing damage.
    """
    name = "Polar Bear"
    description = "Large beast, unaligned"
    challenge_rating = 2
    armor_class = 12
    skills = "Perception +3"
    senses = "Passive Perception 13"
    languages = ""
    strength = Ability(20)
    dexterity = Ability(10)
    constitution = Ability(16)
    intelligence = Ability(2)
    wisdom = Ability(13)
    charisma = Ability(7)
    speed = 40
    swim_speed = 30
    fly_speed = 0
    climb_speed = 0
    hp_max = 42
    hit_dice = "5d10"
    spells = []


class Pony(Monster):
    """
    Hooves.
      Melee Weapon Attack: +4 to hit, reach 5 ft., one target. Hit: 7 (2d4 +
      2) bludgeoning damage.
    """
    name = "Pony"
    description = "Medium beast, unaligned"
    challenge_rating = 0.125
    armor_class = 10
    skills = ""
    senses = "Passive Perception 10"
    languages = ""
    strength = Ability(15)
    dexterity = Ability(10)
    constitution = Ability(13)
    intelligence = Ability(2)
    wisdom = Ability(11)
    charisma = Ability(7)
    speed = 40
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 11
    hit_dice = "2d8"
    spells = []


class Priest(Monster):
    """
    Divine Eminence.
      As a bonus action, the priest can expend a spell slot to cause its
      melee weapon attacks to magically deal an extra 10 (3d6) radiant
      damage to a target on a hit. This benefit lasts until the end of the
      turn. If the priest expends a spell slot of 2nd level or higher, the
      extra damage increases by 1d6 for each level above 1st.
    Spellcasting.
      The priest is a 5th-level spellcaster. Its spellcasting ability is
      Wisdom (spell save DC 13, +5 to hit with spell attacks). The priest
      has the following cleric spells prepared:
      
  
      
      - Cantrips (at will): light, sacred flame, thaumaturgy
      
      - 1st level (4 slots): cure wounds, guiding bolt, sanctuary
      
      - 2nd level (3 slots): lesser restoration, spiritual weapon
      
      - 3rd level (2 slots): dispel magic, spirit guardians
    Mace.
      Melee Weapon Attack: +2 to hit, reach 5 ft., one target. Hit: 3 (1d6)
      bludgeoning damage.
    """
    name = "Priest"
    description = "Medium humanoid, any alignment"
    challenge_rating = 2
    armor_class = 13
    skills = "Medicine +7, Persuasion +3, Religion +4"
    senses = "Passive Perception 13"
    languages = "any two languages"
    strength = Ability(10)
    dexterity = Ability(10)
    constitution = Ability(12)
    intelligence = Ability(13)
    wisdom = Ability(16)
    charisma = Ability(13)
    speed = 25
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 27
    hit_dice = "5d8"
    spells = ["light", "sacred flame", "thaumaturgy", "cure wounds", "guiding bolt", "sanctuary", "lesser restoration", "spiritual weapon", "dispel magic", "spirit guardians"]


class Pseudodragon(Monster):
    """
    Keen Senses.
      The pseudodragon has advantage on Wisdom (Perception) checks that rely
      on sight, hearing, or smell.
    Magic Resistance.
      The pseudodragon has advantage on saving throws against spells and
      other magical effects.
    Limited Telepathy.
      The pseudodragon can magically communicate simple ideas, emotions, and
      images telepathically with any creature within 100 ft. of it that can
      understand a language.
    Variant: Familiar.
      The pseudodragon can serve another creature as a familiar, forming a
      magic, telepathic bond with that willing companion. While the two are
      bonded, the companion can sense what the pseudodragon senses as long
      as they are within 1 mile of each other. While the pseudodragon is
      within 10 feet of its companion, the companion shares the
      pseudodragon's Magic Resistance trait. At any time and for any reason,
      the pseudodragon can end its service as a familiar, ending the
      telepathic bond.
    Bite.
      Melee Weapon Attack: +4 to hit, reach 5 ft., one target. Hit: 4 (1d4 +
      2) piercing damage.
    Sting.
      Melee Weapon Attack: +4 to hit, reach 5 ft., one creature. Hit: 4 (1d4
      + 2) piercing damage, and the target must succeed on a DC 11
      Constitution saving throw or become poisoned for 1 hour. If the saving
      throw fails by 5 or more, the target falls unconscious for the same
      duration, or until it takes damage or another creature uses an action
      to shake it awake.
    """
    name = "Pseudodragon"
    description = "Tiny dragon, neutral good"
    challenge_rating = 0.25
    armor_class = 13
    skills = "Perception +3, Stealth +4"
    senses = "Blindsight 10 ft., Darkvision 60 ft., Passive Perception 13"
    languages = "understands Common and Draconic but can't speak"
    strength = Ability(6)
    dexterity = Ability(15)
    constitution = Ability(13)
    intelligence = Ability(10)
    wisdom = Ability(12)
    charisma = Ability(10)
    speed = 15
    swim_speed = 0
    fly_speed = 60
    climb_speed = 0
    hp_max = 7
    hit_dice = "2d4"
    spells = []


class PurpleWorm(Monster):
    """
    Tunneler.
      The worm can burrow through solid rock at half its burrow speed and
      leaves a 10-foot-diameter tunnel in its wake.
    Multiattack.
      The worm makes two attacks: one with its bite and one with its
      stinger.
    Bite.
      Melee Weapon Attack: +9 to hit, reach 10 ft., one target. Hit: 22 (3d8
      + 9) piercing damage. If the target is a Large or smaller creature, it
      must succeed on a DC 19 Dexterity saving throw or be swallowed by the
      worm. A swallowed creature is blinded and restrained, it has total
      cover against attacks and other effects outside the worm, and it takes
      21 (6d6) acid damage at the start of each of the worm's turns.
      
      If the worm takes 30 damage or more on a single turn from a creature
      inside it, the worm must succeed on a DC 21 Constitution saving throw
      at the end of that turn or regurgitate all swallowed creatures, which
      fall prone in a space within 10 feet of the worm. If the worm dies, a
      swallowed creature is no longer restrained by it and can escape from
      the corpse by using 20 feet of movement, exiting prone.
    Tail Stinger.
      Melee Weapon Attack: +9 to hit, reach 10 ft., one creature. Hit: 19
      (3d6 + 9) piercing damage, and the target must make a DC 19
      Constitution saving throw, taking 42 (12d6) poison damage on a failed
      save, or half as much damage on a successful one.
    """
    name = "Purple Worm"
    description = "Gargantuan monstrosity, unaligned"
    challenge_rating = 15
    armor_class = 18
    skills = ""
    senses = "Blindsight 30 ft., Tremorsense 60 ft., Passive Perception 9"
    languages = ""
    strength = Ability(28)
    dexterity = Ability(7)
    constitution = Ability(22)
    intelligence = Ability(1)
    wisdom = Ability(8)
    charisma = Ability(4)
    speed = 50
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 247
    hit_dice = "15d20"
    spells = []



from dungeonsheets.spells.spells import Spell


class HailOfThorns(Spell):
    """The next time you hit a creature with a ranged weapon attack before the spell
    ends, this spell creates a rain of thorns that sprouts from your ranged weapon
    or ammunition. In addition to the normal effect of the attack, the target of the
    attack and each creature within 5 feet of it must make a Dexterity saving
    throw. A creature takes 1d10 piercing damage on a failed save, or half as much
    damage on a successful one.

    **At Higher Levels:** If you cast this spell using a
    spell slot of 2nd level or higher, the damage increases by 1d10 for each slot
    level above 1st (to a maximum of 6d10).
    """

    name = "Hail Of Thorns"
    level = 1
    casting_time = "1 bonus action"
    casting_range = "Self"
    components = ("V",)
    materials = ""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Conjuration"
    classes = ("Ranger",)


class Hallow(Spell):
    """You touch a point and infuse an area around it with holy (or unholy) power. The
    area can have a radius up to 60 feet, and the spell fails if the radius includes
    an area already under the effect a hallow spell. The affected area is subject
    to the following effects.

    First, celestials, elementals, fey, fiends, and
    undead can't enter the area, nor can such creatures charm, frighten, or possess
    creatures within it. Any creature charmed, frightened, or possessed by such a
    creature is no longer charmed, frightened, or possessed upon entering the area.
    You can exclude one or more of those types of creatures from this effect.


    Second, you can bind an extra effect to the area. Choose the effect from the
    following list, or choose an effect offered by the DM. Som e of these effects
    apply to creatures in the area; you can designate whether the effect applies to
    all creatures, creatures that follow a specific deity or leader, or creatures of
    a specific sort, such as ores or trolls. When a creature that would be affected
    enters the spell's area for the first time on a turn or starts its turn there,
    it can make a Charisma saving throw. On a success, the creature ignores the
    extra effect until it leaves the area.

    Courage
    Affected creatures can't be
    frightened while in the area.

    Darkness
    Darkness fills the area. Normal light,
    as well as magical light created by spells of a lower level than the slot you
    used to cast this spell, can't illuminate the area.

    Daylight
    Bright light fills
    the area. Magical darkness created by spells of a lower level than the slot you
    used to cast this spell can't extinguish the light.

    Energy Protection
    Affected
    creatures in the area have resistance to one damage type of your choice, except
    for bludgeoning, piercing, or slashing.

    Energy Vulnerability
    Affected
    creatures in the area have vulnerability to one damage type of your choice,
    except for bludgeoning, piercing, or slashing.

    Everlasting Rest
    Dead bodies
    interred in the area can't be turned into undead.

    Extradimensional Interference

    Affected creatures can't move or travel using teleportation or by
    extradimensional or interplanar means.

    Fear
    Affected creatures are frightened
    while in the area.

    Silence
    No sound can emanate from within the area, and no
    sound can reach into it.

    Tongues
    Affected creatures can communicate with any
    other creature in the area, even if they don't share a common language.
    """

    name = "Hallow"
    level = 5
    casting_time = "24 hours"
    casting_range = "Touch"
    components = ("V", "S", "M")
    materials = (
        """Herbs, oils, and incense worth at least 1,000 gp, which the spell consumes"""
    )
    duration = "Until dispelled"
    ritual = False
    magic_school = "Evocation"
    classes = ("Cleric",)


class HallucinatoryTerrain(Spell):
    """You make natural terrain in a 150-foot cube in range look, sound, and smell like
    some other sort of natural terrain. Thus, open fields or a road can be made to
    resemble a swamp, hill, crevasse, or some other difficult or impassable terrain.
    A pond can be made to seem like a grassy meadow, a precipice like a gentle
    slope, or a rock-strewn gully like a wide and smooth road. Manufactured
    structures, equipment, and creatures within the area aren't changed in
    appearance.

    The tactile characteristics of the terrain are unchanged, so
    creatures entering the area are likely to see through the illusion. If the
    difference isn't obvious by touch, a creature carefully examining the illusion
    can attempt an Intelligence (Investigation) check against your spell save DC to
    disbelieve it. A creature who discerns the illusion for what it is, sees it as a
    vague image superimposed on the terrain.
    """

    name = "Hallucinatory Terrain"
    level = 4
    casting_time = "10 minutes"
    casting_range = "300 feet"
    components = ("V", "S", "M")
    materials = "A stone, a twig, and a bit of green plant"
    duration = "24 hours"
    ritual = False
    magic_school = "Illusion"
    classes = ("Bard", "Druid", "Warlock", "Wizard")


class Harm(Spell):
    """You unleash a virulent disease on a creature that you can see within range.
    The
    target must make a Constitution saving throw. On a failed save, it takes 14d6
    necrotic damage, or half as much damage on a successful save. The damage can't
    reduce the target's hit points below 1. If the target fails the saving throw,
    its hit point maximum is reduced for 1 hour by an amount equal to the necrotic
    damage it took. Any effect that removes a disease allows a creature's hit point
    maximum to return to normal before that time passes.
    """

    name = "Harm"
    level = 6
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ("V", "S")
    materials = ""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Necromancy"
    classes = ("Cleric",)


class Haste(Spell):
    """Choose a willing creature that you can see within range. Until the spell ends,
    the target's speed is doubled, it gains a +2 bonus to AC, it has advantage on
    Dexterity saving throws, and it gains an additional action on each of its turns.
    That action can be used only to take the Attack (one weapon attack only), Dash,
    Disengage, Hide, or Use an Object action.

    When the spell ends, the target
    can't move or take actions until after its next turn, as a wave of lethargy
    sweeps over it.
    """

    name = "Haste"
    level = 3
    casting_time = "1 action"
    casting_range = "30 feet"
    components = ("V", "S", "M")
    materials = "A shaving of licorice root"
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Transmutation"
    classes = ("Sorcerer", "Wizard")


class Heal(Spell):
    """Choose a creature that you can see within range. A surge of positive energy
    washes through the creature, causing it to regain 70 hit points. The spell also
    ends blindness, deafness, and any diseases affecting the target. This spell has
    no effect on constructs or undead.

    **At Higher Levels:** When you cast this spell
    using aspell slot of 7th level or higher, the amount of healing increases by 10
    for each slot level above 6th.
    """

    name = "Heal"
    level = 6
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ("V", "S")
    materials = ""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Evocation"
    classes = ("Cleric", "Druid")


class HealingSpirit(Spell):
    """You call forth a nature spirit to soothe the wounded. The intangible spirit
    appears in a space that is a 5-foot cube you can see within range. The spirit
    looks like a transparent beast or fey (your choice). Until the spell ends,
    whenever you or a creature you can see moves into the spirits space for the
    first time on a turn or starts its turn there, you can cause the spirit to
    restore ld6 hit points to that creature (no action required). The spirit can't
    heal constructs or undead. As a bonus action on your turn, you can move the
    Spirit up to 30 feet to a space you can see.

    **At Higher Levels:** When you cast
    this spell using a spell slot of 3rd level or higher, the healing increases 1d6
    for each slot level above 2nd.
    """

    name = "Healing Spirit"
    level = 2
    casting_time = "1 bonus action"
    casting_range = "60 feet"
    components = ("V", "S")
    materials = ""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Conjuration"
    classes = ("Druid", "Ranger")


class HealingWord(Spell):
    """A creature of your choice that you can see within range regains hit points equal
    to 1d4 + your spellcasting ability modifier.
    This spell has no effect on undead
    or constructs.

    **At Higher Levels:** When you cast this spell using a spell slot
    of 2nd level or higher, the healing increases by 1d4 for each slot level above
    1st.
    """

    name = "Healing Word"
    level = 1
    casting_time = "1 bonus action"
    casting_range = "60 feet"
    components = ("V",)
    materials = ""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Evocation"
    classes = ("Bard", "Cleric", "Druid")


class HeatMetal(Spell):
    """Choose a manufactured metal object, such as a metal weapon or a suit of heavy or
    medium metal armor, that you can see within range. You cause the object to glow
    red-hot. Any creature in physical contact with the object takes 2d8 fire damage
    when you cast the spell. Until the spell ends, you can use a bonus action on
    each of your subsequent turns to cause this damage again.

    If a creature is
    holding or wearing the object and takes the damage from it, the creature must
    succeed on a Constitution saving throw or drop the object if it can. If it
    doesn't drop the object, it has disadvantage on attack rolls and ability checks
    until the start of your next turn.

    **At Higher Levels:** When you cast this spell
    using a spell slot of 3rd level or higher, the damage increases by 1d8 for each
    slot level above 2nd.
    """

    name = "Heat Metal"
    level = 2
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ("V", "S", "M")
    materials = "A piece of iron and a flame"
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Transmutation"
    classes = ("Bard", "Druid")


class HellishRebuke(Spell):
    """Reaction: you are being damaged by a creature within 60 feet of you that you can
    see.

    You point your finger, and the creature that damaged you is momentarily
    surrounded by hellish flames. The creature must make a Dexterity saving throw.
    It takes 2d10 fire damage on a failed save, or half as much damage on a
    successful one.

    **At Higher Levels:** When you cast this spell using a spell slot
    of 2nd level or higher, the damage increases by 1d10 for each slot level above
    1st.
    """

    name = "Hellish Rebuke"
    level = 1
    casting_time = "Special"
    casting_range = "60 feet"
    components = ("V", "S")
    materials = ""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Evocation"
    classes = ("Warlock",)


class HeroesFeast(Spell):
    """You bring forth a great feast, including magnificent food and drink. The feast
    takes 1 hour to consume and disappears at the end of that time, and the
    beneficial effects don't set in until this hour is over. Up to twelve other
    creatures can partake of the feast.

    A creature that partakes of the feast gains
    several benefits. The creature is cured of all diseases and poison, becomes
    immune to poison and being frightened, and makes all Wisdom saving throws with
    advantage. Its hit point maximum also increases by 2d10, and it gains the same
    number of hit points. These benefits last for 24 hours.
    """

    name = "Heroes Feast"
    level = 6
    casting_time = "10 minutes"
    casting_range = "30 feet"
    components = ("V", "S", "M")
    materials = (
        """A gem-encrusted bowl worth at least 1,000 gp, which the spell consumes"""
    )
    duration = "Instantaneous"
    ritual = False
    magic_school = "Conjuration"
    classes = ("Cleric", "Druid")


class Heroism(Spell):
    """A willing creature you touch is imbued with bravery.
    Until the spell ends, the
    creature is immune to being frightened and gains temporary hit points equal to
    your spellcasting ability modifier at the start of each of its turns. When the
    spell ends, the target loses any remaining temporary hit points from this spell.


    **At Higher Levels:** When you cast this spell using a spell slot of 2nd level or
    higher, you can target one additional creature for each slot level above 1st.
    """

    name = "Heroism"
    level = 1
    casting_time = "1 action"
    casting_range = "Touch"
    components = ("V", "S")
    materials = ""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Enchantment"
    classes = ("Bard", "Paladin")


class Hex(Spell):
    """You place a curse on a creature that you can see within range. Until the spell
    ends, you deal an extra 1d6 necrotic damage to the target whenever you hit it
    with an attack. Also, choose one ability when you cast the spell. The target has
    disadvantage on ability checks made with the chosen ability.

    If the target
    drops to 0 hit points before this spell ends, you can use a bonus action on a
    subsequent turn of yours to curse a new creature.

    A remove curse cast on the
    target ends this spell early.

    **At Higher Levels:** When you cast this spell using
    a spell slot of 3rd or 4th level, you can maintain your concentration on the
    spell for up to 8 hours.
    When you use a spell slot of 5th level or higher, you
    can maintain your concentration on the spell for up to 24 hours.
    """

    name = "Hex"
    level = 1
    casting_time = "1 bonus action"
    casting_range = "90 feet"
    components = ("V", "S", "M")
    materials = "The petrified eye of a newt"
    duration = "Concentration, up to 1 hour"
    ritual = False
    magic_school = "Enchantment"
    classes = ("Warlock",)


class HoldMonster(Spell):
    """Choose a creature that you can see within range. The target must succeed on a
    Wisdom saving throw or be paralyzed for the duration. This spell has no effect
    on undead. At the end of each of its turns, the target can make another Wisdom
    saving throw. On a success, the spell ends on the target.

    At Higher Levels:
    When you cast this spell using a spell slot of 6th level or higher, you can
    target one additional creature for each slot level above 5th. The creatures must
    be within 30 feet of each other when you target them.
    """

    name = "Hold Monster"
    level = 5
    casting_time = "1 action"
    casting_range = "90 feet"
    components = ("V", "S", "M")
    materials = "A small, straight piece of iron"
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Enchantment"
    classes = ("Bard", "Sorcerer", "Warlock", "Wizard")


class HoldPerson(Spell):
    """Choose a humanoid that you can see within range. The target must succeed on a
    Wisdom saving throw or be paralyzed for the duration. At the end of each of its
    turns, the target can make another Wisdom saving throw. On a success, the spell
    ends on the target.

    **At Higher Levels:** When you cast this spell using a spell
    slot of 3rd level or higher, you can target one additional humanoid for each
    slot level above 2nd. The humanoids must be within 30 feet of each other when
    you target them.
    """

    name = "Hold Person"
    level = 2
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ("V", "S", "M")
    materials = "A small, straight piece of iron"
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Enchantment"
    classes = ("Bard", "Cleric", "Druid", "Sorcerer", "Warlock", "Wizard")


class HolyAura(Spell):
    """Divine light washes out from you and coalesces in a soft radiance in a 30-foot
    radius around you.
    Creatures of your choice in that radius when you cast this
    spell shed dim light in a 5-foot radius and have advantage on all saving throws,
    and other creatures have disadvantage on attack rolls against them until the
    spell ends. In addition, when a fiend or an undead hits an affected creature
    with a melee attack, the aura flashes with brilliant light. The attacker must
    succeed on a Constitution saving throw or be blinded until the spell ends.
    """

    name = "Holy Aura"
    level = 8
    casting_time = "1 action"
    casting_range = "Self"
    components = ("V", "S", "M")
    materials = (
        "A tiny reliquary worth at least 1,000 gp containing a sacred relic, such as a"
        " scrap of cloth from a saint's robe or a piece of parchment from a religious"
        " text"
    )
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Abjuration"
    classes = ("Cleric",)


class HolyWeapon(Spell):
    """You imbue a weapon you touch with holy power. Until the spell ends, the weapon
    emits bright light in a 30-foot radius and dim light for an additional 30 feet.
    In addition, weapon attacks made with it deal an extra 2d8 radiant damage on a
    hit. If the weapon isn't already a magic weapon, it becomes one for the
    duration. As a bonus action on your turn, you can dismiss this spell and cause
    the weapon to emit a burst of radiance. Each creature of your choice that you
    can see within 30 feet ofyou must make a Constitution saving throw. On a failed
    save, a creature takes 4d8 radiant damage, and it is blinded for 1 minute. On a
    successful save, a creature takes half as much damage and isn't blinded. At the
    end of each Ofits turns, a blinded creature can make a Constitution saving
    throw, ending the effect on itselfon a success.
    """

    name = "Holy Weapon"
    level = 5
    casting_time = "1 bonus action"
    casting_range = "Touch"
    components = ("V", "S")
    materials = ""
    duration = "Concentration, up to 1 hour"
    ritual = False
    magic_school = "Evocation"
    classes = ("Cleric", "Paladin")


class HungerOfHadar(Spell):
    """You open a gateway to the dark between the stars, a region infested with unknown
    horrors. A 20-foot-radius sphere of blackness and bitter cold appears, centered
    on a point with range and lasting for the duration. This void is filled with a
    cacophony of soft whispers and slurping noises that can be heard up to 30 feet
    away. No light, magical or otherwise, can illuminate the area, and creatures
    fully within the area are blinded.

    The void creates a warp in the fabric of
    space, and the area is difficult terrain. Any creature that starts its turn in
    the area takes 2d6 cold damage. Any creature that ends its turn in the area must
    succeed on a Dexterity saving throw or take 2d6 acid damage as milky,
    otherwordly tentacles rub against it.
    """

    name = "Hunger Of Hadar"
    level = 3
    casting_time = "1 action"
    casting_range = "150 feet"
    components = ("V", "S", "M")
    materials = "A pickled octopus tentacle"
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Conjuration"
    classes = ("Warlock",)


class HuntersMark(Spell):
    """You choose a creature you can see within range and mystically mark it as your
    quarry.
    Until the spell ends, you deal an extra 1d6 damage to the target
    whenever you hit it with a weapon attack, and you have advantage on any Wisdom
    (Perception) or Wisdom (Survival) check you make to find it. If the target drops
    to 0 hit points before this spell ends, you can use a bonus action on a
    subsequent turn of yours to mark a new creature.

    **At Higher Levels:** When you
    cast this spell using a spell slot of 3rd or 4th level, you can maintain your
    concentration on the spell for up to 8 hours.
    When you use a spell slot of 5th
    level or higher, you can maintain your concentration on the spell for up to 24
    hours.
    """

    name = "Hunters Mark"
    level = 1
    casting_time = "1 bonus action"
    casting_range = "90 feet"
    components = ("V",)
    materials = ""
    duration = "Concentration, up to 1 hour"
    ritual = False
    magic_school = "Divination"
    classes = ("Ranger",)


class HypnoticPattern(Spell):
    """You create a twisting pattern of colors that weaves through the air inside a
    30-foot cube within range.
    The pattern appears for a moment and vanishes. Each
    creature in the area who sees the pattern must make a Wisdom saving throw. On a
    failed save, the creature becomes charmed for the duration. While charmed by
    this spell, the creature is incapacitated and has a speed of 0.

    The spell ends
    for an affected creature if it takes any damage or if someone else uses an
    action to shake the creature out of its stupor.
    """

    name = "Hypnotic Pattern"
    level = 3
    casting_time = "1 action"
    casting_range = "120 feet"
    components = ("S", "M")
    materials = (
        "A glowing stick of incense or a crystal vial filled with phosphorescent"
        " material"
    )
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Illusion"
    classes = ("Bard", "Sorcerer", "Warlock", "Wizard")



"""
A collection of monsters from D&D 5e's SRD

This file was autogenerated by https://github.com/stravajiaxen/5e-srd-to-py
"""

from dungeonsheets.monsters.monsters import Monster
from dungeonsheets.stats import Ability


class Xorn(Monster):
    """
    Earth Glide.
      The xorn can burrow through nonmagical, unworked earth and stone.
      While doing so, the xorn doesn't disturb the material it moves
      through.
    Stone Camouflage.
      The xorn has advantage on Dexterity (Stealth) checks made to hide in
      rocky terrain.
    Treasure Sense.
      The xorn can pinpoint, by scent, the location of precious metals and
      stones, such as coins and gems, within 60 ft. of it.
    Multiattack.
      The xorn makes three claw attacks and one bite attack.
    Bite.
      Melee Weapon Attack: +6 to hit, reach 5 ft., one target. Hit: 13 (3d6
      + 3) piercing damage.
    Claw.
      Melee Weapon Attack: +6 to hit, reach 5 ft., one target. Hit: 6 (1d6 +
      3) slashing damage.
    """
    name = "Xorn"
    description = "Medium elemental, neutral"
    challenge_rating = 5
    armor_class = 19
    skills = "Perception +6, Stealth +3"
    senses = "Darkvision 60 ft., Tremorsense 60 ft., Passive Perception 16"
    languages = "Terran"
    strength = Ability(17)
    dexterity = Ability(10)
    constitution = Ability(22)
    intelligence = Ability(11)
    wisdom = Ability(10)
    charisma = Ability(11)
    speed = 20
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 73
    hit_dice = "7d8"
    spells = []



from dungeonsheets.spells.spells import Spell


class DancingLights(Spell):
    """You create up to four torch-sized lights within range, making them appear as
    torches, lanterns, or glowing orbs that hover in the air for the duration.
    You
    can also combine the four lights into one glowing vaguely humanoid form of
    Medium size. Whichever form you choose, each light sheds dim light in a 10-foot
    radius.

    As a bonus action on your turn, you can move the lights up to 60 feet
    to a new spot within range. A light must be within 20 feet of another light
    created by this spell, and a light winks out if it exceeds the spell's range.
    """

    name = "Dancing Lights"
    level = 0
    casting_time = "1 action"
    casting_range = "120 feet"
    components = ("V", "S", "M")
    materials = "A bit of phosphorus or wychwood, or a glowworm"
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Evocation"
    classes = ("Bard", "Sorcerer", "Wizard")


class DanseMacabre(Spell):
    """Threads of dark power leap from your fingers to pierce up to five Small or
    Medium corpses you can see within range. Each corpse immediately stands up and
    becomes undead. You decide whether it is a zombie or a skeleton (the statistics
    for zombies and skeletons are in the Monster Manual), and it gains a bonus to
    its attack and damage rolls equal to your spellcasting ability modifier. You can
    use a bonus action to mentally command the creatures you make with this spell,
    issuing the same command to all of them. To receive the command, a creature must
    be within 60 feet of you. You decide what action the creatures will take and
    where they will move during their next turn, or you can issue a general command,
    such as to guard a chamber or passageway against your foes. lfyou issue no
    commands, the creatures do nothing except defend themselves against hostile
    creatures. Once given an order, the creatures continue to follow it until their
    task is complete.
    The creatures are under your control until the spell ends,
    after which they become inanimate once more.

    **At Higher Levels:** When you cast
    this spell using a spell slot‘ of 6th level or higher, you animate up to two
    additional corpses for each slot level above 5th.
    """

    name = "Danse Macabre"
    level = 5
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ("V", "S")
    materials = ""
    duration = "Concentration, up to 1 hour"
    ritual = False
    magic_school = "Necromancy"
    classes = ("Warlock", "Wizard")


class Darkness(Spell):
    """Magical darkness spreads from a point you choose within range to fill a 15-foot
    radius sphere for the duration.
    The darkness spreads around corners. A creature
    with darkvision can't see through this darkness, and nonmagical light can't
    illuminate it.

    If the point you choose is on an object you are holding or one
    that isn't being worn or carried, the darkness emanates from the object and
    moves with it. Completely covering the source of the darkness with an opaque
    object, such as a bowl or a helm, blocks the darkness.

    If any of this spell's
    area overlaps with an area of light created by a spell of 2nd level or lower,
    the spell that created the light is dispelled.
    """

    name = "Darkness"
    level = 2
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ("V", "M")
    materials = "Bat fur and a drop of pitch or piece of coal"
    duration = "Concentration, up to 10 minutes"
    ritual = False
    magic_school = "Evocation"
    classes = ("Sorcerer", "Warlock", "Wizard")


class Darkvision(Spell):
    """You touch a willing creature to grant it the ability to see in the dark.
    For the
    duration, that creature has darkvision out to a range of 60 feet.
    """

    name = "Darkvision"
    level = 2
    casting_time = "1 action"
    casting_range = "Touch"
    components = ("V", "S", "M")
    materials = "Either a pinch of dried carrot or an agate"
    duration = "8 hours"
    ritual = False
    magic_school = "Transmutation"
    classes = ("Sorcerer", "Wizard", "Druid", "Ranger")


class Dawn(Spell):
    """The light of dawn shines down on a location you specify within range. Until the
    spell ends, a 30-foot-radius.40-foot-high cylinder of bright light glimmers
    there. This light is sunlight. When the cylinder appears, each creature in it
    must make a Constitution saving throw, taking 4d10 radiant damage on a failed
    save, or half as much damage on a successful one. A creature must also make this
    saving throw whenever it ends its turn in the cylinder. If you're within 60
    feet of the cylinder, you can move it up to 60 feet as a bonus action on your
    turn.
    """

    name = "Dawn"
    level = 5
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ("V", "S", "M")
    materials = "A sunburst pendant worth at least 100 gp"
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Evocation"
    classes = ("Cleric", "Wizard")


class Daylight(Spell):
    """A 60-foot-radius sphere of light spreads out from a point you choose within
    range.
    The sphere is bright light and sheds dim light for an additional 60
    feet.

    If you chose a point on an object you are holding or one that isn't
    being worn or carried, the light shines from the object with and moves with it.
    Completely covering the affected object with an opaque object, such as a bowl or
    a helm, blocks the light.

    If any of this spell's area overlaps with an area
    of darkness created by a spell of 3rd level or lower, the spell that created the
    darkness is dispelled.
    """

    name = "Daylight"
    level = 3
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ("V", "S")
    materials = ""
    duration = "1 hour"
    ritual = False
    magic_school = "Evocation"
    classes = ("Cleric", "Druid", "Paladin", "Ranger", "Sorcerer")


class DeathWard(Spell):
    """You touch a creature and grant it a measure of protection from death.
    The first
    time the target would drop to 0 hit points as a result of taking damage, the
    target instead drops to 1 hit point, and the spell ends. If the spell is still
    in effect when the target is subjected to an effect that would kill it
    instantaneously without dealing damage, that effect is instead negated against
    the target, and the spells ends.
    """

    name = "Death Ward"
    level = 4
    casting_time = "1 action"
    casting_range = "Touch"
    components = ("V", "S")
    materials = ""
    duration = "8 hours"
    ritual = False
    magic_school = "Abjuration"
    classes = ("Cleric", "Paladin")


class DelayedBlastFireball(Spell):
    """A beam of yellow light flashes from your pointing finger, then condenses to
    linger at a chosen point within range as a glowing bead for the duration.
    When
    the spell ends, either because your concentration is broken or because you
    decide to end it, the bead blossoms with a low roar into an explosion of flame
    that spreads around corners. Each creature in a 20-foot-radius sphere centered
    on that point must make a Dexterity saving throw. A creature takes fire damage
    equal to the total accumulated damage on a failed save, or half as much damage
    on a successful one.

    The spell's base damage is 12d6. If at the end of your
    turn the bead has not yet detonated, the damage increases by 1d6.

    If the
    glowing bead is touched before the interval has expired, the creature touching
    it must make a Dexterity saving throw. On a failed save, the spell ends
    immediately, causing the bead to erupt in flame. On a successful save, the
    creature can throw the bead up to 40 feet. When it strikes a creature or a solid
    object, the spell ends, and the bead explodes.
    The fire damages objects in the
    area and ignites flammable objects that aren't being worn or carried.

    At
    Higher Levels: When you cast this spell using a spell slot of 8th level or
    higher, the base damage increases by 1d6 for each slot level above 7th.
    """

    name = "Delayed Blast Fireball"
    level = 7
    casting_time = "1 action"
    casting_range = "150 feet"
    components = ("V", "S", "M")
    materials = "A tiny ball of bat guano and sulfur"
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Evocation"
    classes = ("Sorcerer", "Wizard")


class Demiplane(Spell):
    """You create a shadowy door on a flat solid surface that you can see within range.

    The door is large enough to allow Medium creatures to pass through unhindered.
    When opened, the door leads to a demiplane that appears to be an empty room 30
    feet in each dimension, made of wood or stone. When the spell ends, the door
    disappears, and any creatures or objects inside the demiplane remain trapped
    there, as the door also disappears from the other side.

    Each time you cast
    this spell, you can create a new demiplane, or have the shadowy door connect to
    a demiplane you created with a previous casting of this spell. Additionally, if
    you know the nature and contents of a demiplane created by a casting of  this
    spell by another creature, you can have the shadowy door connect to its
    demiplane instead.
    """

    name = "Demiplane"
    level = 8
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ("S",)
    materials = ""
    duration = "1 hour"
    ritual = False
    magic_school = "Conjuration"
    classes = ("Warlock", "Wizard")


class DestructiveWave(Spell):
    """You strike the ground, creating a burst of divine energy that ripples outward
    from you. Each creature you choose within 30 feet of you must succeed on a
    Constitution saving throw or take 5d6 thunder damage, as well as 5d6 radiant or
    necrotic damage (your choice), and be knocked prone. A creature that succeeds on
    its saving throw takes half as much damage and isn't knocked prone.
    """

    name = "Destructive Wave"
    level = 5
    casting_time = "1 action"
    casting_range = "Self (30-foot radius)"
    components = ("V",)
    materials = ""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Evocation"
    classes = ("Paladin",)


class DetectEvilAndGood(Spell):
    """For the duration, you know if there is an aberration, celestial, elemental, fey,
    fiend, or undead within 30 feet of you, as well as where the creature is
    located. Similarly, you know if there is a place or object within 30 feet of you
    that has been magically consecrated or desecrated.

    The spell can penetrate
    most barriers, but it is blocked by 1 foot of stone, 1 inch of common metal, a
    thin sheet of lead, or 3 feet of wood or dirt.
    """

    name = "Detect Evil And Good"
    level = 1
    casting_time = "1 action"
    casting_range = "Self"
    components = ("V", "S")
    materials = ""
    duration = "Concentration, up to 10 minutes"
    ritual = False
    magic_school = "Divination"
    classes = ("Cleric", "Paladin")


class DetectMagic(Spell):
    """For the duration, you sense the presence of magic within 30 feet of you. If you
    sense magic in this way, you can use your action to see a faint aura around any
    visible creature or object in the area that bears magic, and you learn its
    school of magic, if any.

    The spell can penetrate most barriers, but is blocked
    by 1 foot of stone, 1 inch of common metal, a thin sheet of lead, or 3 feet of
    wood or dirt.
    """

    name = "Detect Magic"
    level = 1
    casting_time = "1 action"
    casting_range = "Self"
    components = ("V", "S")
    materials = ""
    duration = "Concentration, up to 10 minutes"
    ritual = True
    magic_school = "Divination"
    classes = ("Bard", "Cleric", "Druid", "Paladin", "Ranger", "Sorcerer", "Wizard")


class DetectPoisonAndDisease(Spell):
    """For the duration, you can sense the presence and location of poisons, poisonous
    creatures, and diseases within 30 feet of you. You also identify the kind of
    poison, poisonous creature, or disease in each case.

    The spell can penetrate
    most barriers, but is blocked by 1 foot of stone, 1 inch of common metal, a thin
    sheet of lead, or 3 feet of wood or dirt.
    """

    name = "Detect Poison And Disease"
    level = 1
    casting_time = "1 action"
    casting_range = "Self"
    components = ("V", "S", "M")
    materials = "A yew leaf"
    duration = "Concentration, up to 10 minutes"
    ritual = False
    magic_school = "Divination"
    classes = ("Cleric", "Druid", "Paladin", "Ranger")


class DetectThoughts(Spell):
    """For the duration, you can read the thoughts of certain creatures.
    When you cast
    the spell and as your action on each turn until the spell ends, you can focus
    your mind on any one creature that you can see within 30 feet of you. If the
    creature you choose has an Intelligence of 3 or lower or doesn't speak any
    language, the creature is unaffected.

    You initially learn the surface thoughts
    of the creature-what is most on its mind in that moment. As an action, you can
    either shift your attention to another creature's thoughts or attempt to probe
    deeper into the same creature's mind. If you probe deeper, the target must make
    a W isdom saving throw. If it fails, you gain insight into its reasoning (if
    any), its emotional state, and something that loom s large in its mind (such as
    something it worries over, loves, or hates). If it succeeds, the spell ends.
    Either way, the target knows that you are probing into its mind, and unless you
    shift your attention to another creature's thoughts, the creature can use its
    action on its turn to make an Intelligence check contested by your Intelligence
    check; if it succeeds, the spell ends.

    Questions verbally directed at the
    target creature naturally shape the course of its thoughts, so this spell is
    particularly effective as part of an interrogation.

    You can also use this
    spell to detect the presence of thinking creatures you can't see. When you cast
    the spell or as your action during the duration, you can search for thoughts
    within 30 feet of you. The spell can penetrate barriers, but 2 feet of rock, 2
    inches of any metal other than lead, or a thin sheet of lead blocks you. You
    can't detect a creature with an Intelligence of 3 or lower or one that doesn't
    speak any language.

    Once you detect the presence of a creature in this way,
    you can read its thoughts for the rest of the duration as described above, even
    if you can't see it, but it must still be within range.
    """

    name = "Detect Thoughts"
    level = 2
    casting_time = "1 action"
    casting_range = "Self"
    components = ("V", "S", "M")
    materials = "A copper piece"
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Divination"
    classes = ("Bard", "Sorcerer", "Wizard")


class DimensionDoor(Spell):
    """You teleport yourself from your current location to any other spot within range.
    You arrive at exactly the spot desired. It can be a place you can see, one you
    can visualize, or one you can describe by stating distance and direction, such
    as "200 feet straight downward" or "upward to the northwest at a 45-degree
    angle, 300 feet".

    You can bring along objects as long as their weight doesn't
    exceed what you can carry. You can also bring one willing creature of your size
    or smaller who is carrying gear up to its carrying capacity. The creature must
    be within 5 feet of you when you cast this spell.

    If you would arrive in a
    place already occupied by an object or a creature, you and any creature
    traveling with you each take 4d6 force damage, and the spell fails to teleport
    you.
    """

    name = "Dimension Door"
    level = 4
    casting_time = "1 action"
    casting_range = "500 feet"
    components = ("V",)
    materials = ""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Conjuration"
    classes = ("Bard", "Sorcerer", "Warlock", "Wizard")


class DisguiseSelf(Spell):
    """You make yourself, including your clothing, armor, weapons, and other
    belongings on your person, look different until the spell ends or until you
    use your action to dismiss it.
    You can seem 1 foot shorter or taller and can
    appear thin, fat, or in between. You can't change your body type, so you must
    adopt a form that has the same basic arrangement of limbs. Otherwise, the extent
    of the illusion is up to you.

    The changes wrought by this spell fail to hold
    up to physical inspection. For example, if you use this spell to add a hat to
    your outfit, objects pass through the hat, and anyone who touches it would feel
    nothing or would feel your head and hair. If you use this spell to appear
    thinner than you are, the hand of som eone who reaches out to touch you would
    bump into you while it was seemingly still in midair. To discern that you are
    disguised, a creature can use its action to inspect your appearance and must
    succeed on an Intelligence (Investigation) check against your spell save DC.
    """

    name = "Disguise Self"
    level = 1
    casting_time = "1 action"
    casting_range = "Self"
    components = ("V", "S")
    materials = ""
    duration = "1 hour"
    ritual = False
    magic_school = "Illusion"
    classes = ("Bard", "Sorcerer", "Wizard")


class Disintegrate(Spell):
    """A thin green ray springs from your pointing finger to a target that you can see
    within range.
    The target can be a creature, an object, or a creation of magical
    force, such as the wall created by wall of force.

    A creature targeted by this
    spell must make a Dexterity saving throw. On a failed save, the target takes
    10d6 + 40 force damage. If this damage reduces the target to 0 hit points, it is
    disintegrated.

    A disintegrated creature and everything it is wearing and
    carrying, except magic items, are reduced to a pile of fine gray dust. The
    creature can be restored to life only by means of a true resurrection or a wish
    spell.

    This spell automatically disintegrates a Large or smaller nonmagical
    object or a creation of magical force. If the target is a Huge or larger object
    or creation of force, this spell disintegrates a 10-foot-cube portion of it. A
    magic item is unaffected by this spell.

    **At Higher Levels:** When you cast this
    spell using a spell slot of 7th level or higher, the damage increases by 3d6 for
    each slot level above 6th.
    """

    name = "Disintegrate"
    level = 6
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ("V", "S", "M")
    materials = "A lodestone and a pinch of dust"
    duration = "Instantaneous"
    ritual = False
    magic_school = "Transmutation"
    classes = ("Sorcerer", "Wizard")


class DispelEvilAndGood(Spell):
    """Shimmering energy surrounds and protects you from fey, undead, and creatures
    originating from beyond the Material Plane. For the duration, celestials,
    elementals, fey, fiends, and undead have disadvantage on attack rolls against
    you.

    You can end the spell early by using either of the following special
    functions.

    Break  Enchantment
    As your action, you touch a creature you can
    reach that is charmed, frightened, or possessed by a celestial, an elemental, a
    fey, a fiend, or an undead. The creature you touch is no longer charmed,
    frightened, or possessed by such creatures.

    Dismissal
    As your action, make a
    melee spell attack against a celestial, an elemental, a fey, a fiend, or an
    undead you can reach. On a hit, you attempt to drive the creature back to its
    home plane. The creature must succeed on a Charisma saving throw or be sent back
    to its home plane (if it isn't there already). If they aren't on their home
    plane, undead are sent to the Shadowfell, and fey are sent to the Feywild.
    """

    name = "Dispel Evil And Good"
    level = 5
    casting_time = "1 action"
    casting_range = "Self"
    components = ("V", "S", "M")
    materials = "Holy water or powdered silver and iron"
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Abjuration"
    classes = ("Cleric", "Paladin")


class DispelMagic(Spell):
    """Choose any creature, object, or magical effect within range. Any spell of 3rd
    level or lower on the target ends. For each spell of 4th level or higher on the
    target, make an ability check using your spellcasting ability. The DC equals 10
    + the spell's level. On a successful check, the spell ends.

    At Higher Levels:
    When you cast this spell using a spell slot of 4th level or higher, you
    automatically end the effects of a spell on the target if the spell's level is
    equal to or less than the level of the spell slot you used.
    """

    name = "Dispel Magic"
    level = 3
    casting_time = "1 action"
    casting_range = "120 feet"
    components = ("V", "S")
    materials = ""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Abjuration"
    classes = ("Bard", "Cleric", "Druid", "Paladin", "Sorcerer", "Warlock", "Wizard")


class DissonantWhispers(Spell):
    """You whisper a discordant melody that only one creature of your choice within
    range can hear, wracking it with terrible pain.
    The target must make a Wisdom
    saving throw. On a failed save, it takes 3d6 psychic damage and must immediately
    use its reaction , if available, to move as far as its speed allows away from
    you. The creature doesn't move into obviously dangerous ground, such as a fire
    or a pit. On a successful save, the target takes half as much damage and doesn't
    have to move away. A deafened creature automatically succeeds on the save.

    At
    Higher Levels: When you cast this spell using a spell slot of 2nd level or
    higher, the damage increases by 1d6 for each slot level above 1st
    """

    name = "Dissonant Whispers"
    level = 1
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ("V",)
    materials = ""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Enchantment"
    classes = ("Bard",)


class Divination(Spell):
    """Your magic and an offering put you in contact with a god or a god's servants.
    You ask a single question concerning a specific goal, event, or activity to
    occur within 7 days. The DM offers a truthful reply. The reply might be a short
    phrase, a cryptic rhyme, or an omen.

    The spell doesn't take into account any
    possible circumstances that might change the outcome, such as the casting of
    additional spells or the loss or gain of a companion.

    If you cast this spell
    two or more times before finishing your next long rest, there is a cumulative 25
    percent chance for each casting after the first that you get a random reading.
    The DM makes this roll in secret.
    """

    name = "Divination"
    level = 4
    casting_time = "1 action"
    casting_range = "Self"
    components = ("V", "S", "M")
    materials = (
        "Incense and a sacrificial offering appropriate to your religion, together"
        " worth at least 25 gp, which the spell consumes"
    )
    duration = "Instantaneous"
    ritual = True
    magic_school = "Divination"
    classes = ("Cleric",)


class DivineFavor(Spell):
    """Your prayer empowers you with divine radiance. Until the spell ends, your weapon
    attacks deal and extra 1d4 radiant damage on a hit.
    """

    name = "Divine Favor"
    level = 1
    casting_time = "1 bonus action"
    casting_range = "Self"
    components = ("V", "S")
    materials = ""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Evocation"
    classes = ("Paladin",)


class DivineWord(Spell):
    """You utter a divine word, imbued with the power that shaped the
    world at the dawn of creation.  Choose any number of creatures you
    can see within range. Each creature that can hear you must make a
    Charisma saving throw. On a failed save, a creature suffers an
    effect based on its current hit points:

    - 50 hit points or fewer: deafened for 1 minute
    - 40 hit points or fewer: deafened and blinded for 10 minutes
    - 30 hit points or fewer: blinded, deafened, and stunned for 1
      hour
    - 20 hit points or fewer: killed instantly

    Regardless of its current hit points, a celestial, an elemental, a
    fey, or a fiend that fails its save is forced back to its plane of
    origin (if it isn't there already) and can't return to your
    current plane for 24 hours by any means short of a wish spell.

    """

    name = "Divine Word"
    level = 7
    casting_time = "1 bonus action"
    casting_range = "30 feet"
    components = ("V",)
    materials = ""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Evocation"
    classes = ("Cleric",)


class DominateBeast(Spell):
    """You attempt to beguile a beast that you can see within range. It must
    succeed on a Wisdom saving throw or be charmed by you for the duration. If
    you or creatures that are friendly to you are fighting it, it has advantage
    on the saving throw.

    While the beast is charmed, you have a telepathic link with it
    as long as the two of you are on the same plane of existence. You can use
    this telepathic link to issue commands to the creature while you are
    conscious (no action required), which it does its best to obey. You can
    specify a simple and general course of action, such as "Attack that
    creature," "Run over there," or "Fetch that object." If the creature
    completes the order and doesn't receive further direction from you, it
    defends and preserves itself to the best of its ability.

    You can use your action to take total and precise control of the target.
    Until the end of your next turn, the creature takes only the actions you
    choose, and doesn't do anything that you don't allow it to do. During this
    time, you can also cause the creature to use a reaction, but this requires
    you to use your own reaction as well.

    Each time the target takes damage, it makes a new Wisdom saving throw
    against the spell. If the saving throw succeeds, the spell ends.

    **At Higher Levels:** When you cast this spell with a 5th-level spell slot,
    the duration is concentration, up to 10 minutes. When you use a 6th-level
    spell slot, the duration is concentration, up to 1 hour. When you use
    a spell slot of 7th level or higher, the duration is concentration, up to
    8 hours
    """

    name = "Dominate Beast"
    level = 4
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ("V", "S")
    materials = ""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Enchantment"
    classes = ("Druid", "Sorcerer")


class DominateMonster(Spell):
    """You attempt to beguile a creature that you can see within range.
    It must
    succeed on a Wisdom saving throw or be charmed by you for the duration. If you
    or creatures that are friendly to you are fighting it, it has advantage on the
    saving throw.

    While the creature is charmed, you have a telepathic link with
    it as long as the two of you are on the same plane of existence. You can use
    this telepathic link to issue commands to the creature while you are conscious
    (no action required), which it does its best to obey. You can specify a simple
    and general course of action, such as "Attack that creature", "Run over
    there", or "Fetch that object". If the creature completes the order and
    doesn't receive further direction from you, it defends and preserves itself to
    the best of its ability.

    You can use your action to take total and precise
    control of the target. Until the end of your next turn, the creature takes only
    the actions you choose, and doesn't do anything that you don't allow it to do.
    During this time, you can also cause the creature to use a reaction, but this
    requires you to use your own reaction as well.

    Each time the target takes
    damage, it makes a new Wisdom saving throw against the spell. If the saving
    throw succeeds, the spell ends.

    **At Higher Levels:** When you cast this spell with
    a 9th-level spell slot, the duration is concentration, up to 8 hours.
    """

    name = "Dominate Monster"
    level = 8
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ("V", "S")
    materials = ""
    duration = "Concentration, up to 1 hour"
    ritual = False
    magic_school = "Enchantment"
    classes = ("Bard", "Sorcerer", "Warlock", "Wizard")


class DominatePerson(Spell):
    """You attempt to beguile a humanoid that you can see within range.
    It must succeed on a Wisdom saving throw or be charmed by you for the
    duration. If you or creatures that are friendly to you are fighting it, it
    has advantage on the saving throw.

    While the target is charmed, you have a telepathic link with it as long as
    the two of you are on the same plane of existence. You can use this
    telepathic link to issue commands to the creature while you are conscious
    (no action required), which it does its best to obey. You can specify
    a simple and general course of action, such as "Attack that creature",
    "Run over there", or "Fetch that object". If the creature completes
    the order and doesn't receive further direction from you, it defends and
    preserves itself to the best of its ability.

    You can use your action to take total and precise control of
    the target. Until the end of your next turn, the creature takes only
    the actions you choose, and doesn't do anything that you don't allow it to
    do. During this time you can also cause the creature to use a reaction, but
    this requires you to use your own reaction as well.

    Each time the target takes damage, it makes a new Wisdom saving throw
    against the spell. If the saving throw succeeds, the spell ends.

    **At Higher Levels:** When you cast this spell using a 6th-level spell slot,
    the duration is concentration, up to 10 minutes. When you use a 7th-level
    spell slot, the duration is concentration, up to 1 hour. When you use
    a spell slot of 8th level or higher, the duration is concentration, up to
    8 hours.
    """

    name = "Dominate Person"
    level = 5
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ("V", "S")
    materials = ""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Enchantment"
    classes = ("Bard", "Sorcerer", "Wizard")


class DragonsBreath(Spell):
    """You touch one willing creature and imbue it with the power to spew magical
    energy from its mouth, provided it has one. Choose acid, cold, fire, lightning,
    or poison. Until the spell ends, the creature can use an action to exhale energy
    of the chosen type in a 15-foot cone. Each creature in that area must make a
    Dexterity saving throw, taking 3d6 damage of the chosen type on a failed save,
    or half as much damage on a successful one.

    **At Higher Levels:** When you cast
    this spell using a spell slot of 3rd level or higher, the damage increases by
    1d6 for each slot level above 2nd.
    """

    name = "Dragons Breath"
    level = 2
    casting_time = "1 bonus action"
    casting_range = "Touch"
    components = ("V", "S", "M")
    materials = "A hot pepper"
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Transmutation"
    classes = ("Sorcerer", "Wizard")


class DrawmijsInstantSummons(Spell):
    """You touch an object weighing 10 pounds or less whose longest dimension is 6 feet
    or less.
    The spell leaves an invisible mark on its surface and invisibly
    inscribes the name of the item on the sapphire you use as the material
    component. Each time you cast this spell, you must use a different sapphire.


    At any time thereafter, you can use your action to speak the item's name and
    crush the sapphire. The item instantly appears in your hand regardless of
    physical or planar distances, and the spell ends. If another creature is holding
    or carrying the item, crushing the sapphire doesn't transport the item to you,
    but instead you learn who the creature possessing the object is and roughly
    where that creature is located at that moment.

    Dispel magic or a similar
    effect successfully applied to the sapphire ends this spell's effect.
    """

    name = "Drawmijs Instant Summons"
    level = 6
    casting_time = "1 minute"
    casting_range = "Touch"
    components = ("V", "S", "M")
    materials = "A sapphire worth 1,000 gp"
    duration = "Until dispelled"
    ritual = True
    magic_school = "Conjuration"
    classes = ("Wizard",)


class Dream(Spell):
    """This spell shapes a creature's dreams. Choose a creature known to you
    as the target of this spell. The target must be on the same plane of
    existence as you. Creatures that don't sleep, such as elves, can't be
    contacted by this spell. You, or a willing creature you touch, enters
    a trance state, acting as a messenger. While in the trance, the
    messenger is aware of his or her surroundings, but can't take actions
    or move.

    If the target is asleep, the messenger appears in the
    target's dreams and can converse with the target as long as it remains
    asleep, through the duration of the spell. The messenger can also
    shape the environment of the dream, creating landscapes, objects, and
    other images. The messenger can emerge from the trance at any time,
    ending the effect of the spell early. The target recalls the dream
    perfectly upon waking. If the target is awake when you cast the spell,
    the messenger knows it, and can either end the trance (and the spell)
    or wait for the target to fall asleep, at which point the messenger
    appears in the target's dreams.

    You can make the messenger appear
    monstrous and terrifying to the target. If you do, the messenger can
    deliver a message of no more than ten words and then the target must
    make a Wisdom saving throw. On a failed save, echoes of the phantasmal
    monstrosity spawn a nightmare that lasts the duration of the target's
    sleep and prevents the target from gaining any benefit from that
    rest. In addition, when the target wakes up, it takes 3d6 psychic
    damage.

    If you have a body part, lock of hair, clipping from a nail,
    or similar portion of the target's body, the target makes its saving
    throw with disadvantage.
    """

    name = "Dream"
    level = 5
    casting_time = "1 minute"
    casting_range = "Special"
    components = ("V", "S", "M")
    materials = """A handful of sand, a dab of ink, and a writing quill plucked from a
sleeping bird"""
    duration = "8 hours"
    ritual = False
    magic_school = "Illusion"
    classes = ("Bard", "Druid", "Warlock", "Wizard")


class DruidGrove(Spell):
    """You invoke the spirits of nature to protect an area outdoors or underground. The
    area can be as small as a 30-foot cube or as large as a 90-foot cube. Buildings
    and other structures are excluded from the affected area. If you cast this
    spell in the same area every day for a year, the spell lasts until dispelled.
    The spell creates the following effects within the area. When you cast this
    spell, you can specify creatures as friends who are immune to the effects. You
    can also specify a password that, when spoken aloud, makes the speaker immune to
    these effects. The entire warded area radiates magic. A dispel magic cast on
    the area, if successful, removes only one of the following effects, not the
    entire area. That spell's caster chooses which effect to end. Only when all its
    effects are gone is this spell dispelled.
    Solid Fog. You can fill any number of
    5-foot squares on the ground with thick fog, making them heavily obscured. The
    fog reaches 10 feet high. In addition, every foot of movement through the fog
    costs 2 extra feet. To a creature immune to this effect, the fog obscures
    nothing and looks like soft mist, with motes of green light floating in the air.

    Grasping Undergrowth. You can fill any number of 5-foot squares on the ground
    that aren't filled with fog with grasping weeds and vines, as if they were
    affected by an entangle spell. To a creature immune to this effect, the weeds
    and vines feel soft and reshape themselves to serve as temporary seats or beds.

    Grove Guardians. You can animate up to four trees in the area, causing them to
    uproot themselves from the ground. These trees have the same statistics as an
    awakened tree, which appears in the Monster Manual, except they can't speak, and
    their bark is covered with druidic symbols. If any creature not immune to this
    effect enters the warded area, the grove guardians fight until they have driven
    off or slain the intruders. The grove guardians also obey your spoken commands
    (no action required by you) that you issue while in the area. Ifyou don't give
    them commands and no intruders are present, the grove guardians do nothing. The
    grove guardians can‘t leave the warded area. When the spell ends, the magic
    animating them disappears, and the trees take root again if possible.
    Additional
    Spell Effect. You can place your choice of one of the following magical effects
    within the warded area:
    - A constant gust of Wind in two locations of your
    choice
    - Spike growth in one location of your choice
    - Wind wall in two
    locations of your choice
    To a creature immune to this effect, the winds are a
    fragrant, gentle breeze, and the area of spike growth is harmless.
    """

    name = "Druid Grove"
    level = 6
    casting_time = "10 minutes"
    casting_range = "Touch"
    components = ("V", "S", "M")
    materials = (
        "Mistletoe, which the spell consumes, that was harvested with a golden sickle"
        " under the light of a full moon"
    )
    duration = "24 hours"
    ritual = False
    magic_school = "Abjuration"
    classes = ("Druid",)


class Druidcraft(Spell):
    """Whispering to the spirits of nature, you create one of the following effects
    within range:

    - You create a tiny, harmless sensory effect that predicts what
      the weather will be at your location for the next 24 hours. The
      effect might manifest as a golden orb for clear skies, a cloud
      for rain, falling snowflakes for snow, and so on. This effect
      persists for 1 round.
    - You instantly make a flower blossom, a seed pod open, or a leaf
      bud bloom.
    - You create an instantaneous, harmless sensory effect, such as
      falling leaves, a puff of wind, the sound of a small animal, or
      the faint odor of skunk. The effect must fit in a 5-foot cube.
    - You instantly light or snuff out a candle, a torch, or a small
      campfire.

    """

    name = "Druidcraft"
    level = 0
    casting_time = "1 action"
    casting_range = "30 feet"
    components = ("V", "S")
    materials = ""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Transmutation"
    classes = ("Druid",)


class DustDevil(Spell):
    """(a pinch of dust)
    Choose an unoccupied 5-foot cube of air that you can see
    within range. An elemental force that resembles a dust devil appears in the cube
    and lasts for the spell's duration.
    Any creature that ends its turn within 5
    feet of the dust devil must make a Strength saving throw. On a failed save, the
    creature takes 1d8 bludgeoning damage and is pushed 10 feet away. On a
    successful save, the creature takes half as much damage and isn't pushed.
    As a
    bonus action, you can move the dust devil up to 30 feet in any direction. If the
    dust devil moves over sand, dust, loose dirt, or small gravel, it sucks up the
    material and forms a 10-foot-radius cloud of debris around itself that lasts
    until the start of your next turn. The cloud heavily obscures its area.
    At
    Higher Levels. When you cast this spell using a spell slot of 3rd level or
    higher, the damage increases by 1d8 for each slot level above 2nd.
    """

    name = "Dust Devil"
    level = 2
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ("V", "S", "M")
    materials = ""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Conjuration"
    classes = ("Druid", "Sorcerer", "Wizard")



import random
import re
from collections import namedtuple
from itertools import groupby

from dungeonsheets.exceptions import DiceError

dice_re = re.compile(r"(\d+)d(\d+)", flags=re.I)
dice_part_re = re.compile(r"[0-9d]+", flags=re.I)
Dice = namedtuple("Dice", ("num", "faces"))


def read_dice_str(dice_str):
    """Interpret a D&D dice string, eg. 3d10.

    Returns
    -------
    dice : tuple
      A named tuple with the scheme (num, faces), so '3d10' return
      (num=3, faces=10)

    """
    match = dice_re.match(dice_str)
    if match is None:
        raise DiceError(f"Cannot interpret dice string {dice_str}")
    dice = Dice(num=int(match.group(1)), faces=int(match.group(2)))
    return dice


def combine_dice(dice_str):
    """Condense a dice string into its simplest representation.

    For example: "1d8 + 5 + 2d8 + 2" would be reduced to "3d8 + 7".

    Returns
    =======
    new_dice_str
      A new, condensed string for the given dice.
    
    """
    dice = []
    bonuses = 0
    # Sort out each portion of the dice string
    for part in dice_part_re.findall(dice_str):
        try:
            dice.append(read_dice_str(part))
        except DiceError:
            bonuses += int(part)
    # Recombine the dice into a more concise string
    new_parts = []
    dice = sorted(dice, key=lambda x: x.faces)
    for faces, grp in groupby(dice, key=lambda x: x.faces):
        new_parts.append(f"{sum([d.num for d in grp])}d{faces}")
    if bonuses > 0:
        new_parts.append(str(bonuses))
    new_dice_str = " + ".join(new_parts)
    return new_dice_str


def roll(a, b=None):
    """roll(20) means roll 1d20, roll(2, 6) means roll 2d6"""
    if b is None:
        return random.randint(1, a)
    else:
        return sum([random.randint(1, b) for _ in range(a)])
